# Copyright (c) 2014 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_log import log as logging
import pecan

from designate.api.v2.controllers import rest
from designate.objects.adapters import DesignateAdapter
from designate.objects import Pool
from designate import utils


LOG = logging.getLogger(__name__)


class PoolsController(rest.RestController):
    SORT_KEYS = ['created_at', 'id', 'updated_at', 'name']

    @pecan.expose(template='json:', content_type='application/json')
    @utils.validate_uuid('pool_id')
    def get_one(self, pool_id):
        """Get the specific Pool"""
        request = pecan.request
        context = request.environ['context']

        pool = self.central_api.get_pool(context, pool_id)

        LOG.info("Retrieved %(pool)s", {'pool': pool})

        return DesignateAdapter.render('API_v2', pool, request=request)

    @pecan.expose(template='json:', content_type='application/json')
    def get_all(self, **params):
        """List all Pools"""
        request = pecan.request
        context = request.environ['context']

        # Extract the pagination params
        marker, limit, sort_key, sort_dir = utils.get_paging_params(
                context, params, self.SORT_KEYS)

        # Extract any filter params.
        accepted_filters = ('name', )
        criterion = self._apply_filter_params(
            params, accepted_filters, {})

        pools = self.central_api.find_pools(
            context, criterion, marker, limit, sort_key, sort_dir)

        LOG.info("Retrieved %(pools)s", {'pools': pools})

        return DesignateAdapter.render('API_v2', pools, request=request)

    @pecan.expose(template='json:', content_type='application/json')
    def post_all(self):
        """Create a Pool"""

        LOG.warning("Use of this API Method is DEPRECATED. This will have "
                    "unforeseen side affects when used with the "
                    "designate-manage pool commands")

        request = pecan.request
        response = pecan.response
        context = request.environ['context']
        body = request.body_dict

        pool = DesignateAdapter.parse('API_v2', body, Pool())

        pool.validate()

        # Create the pool
        pool = self.central_api.create_pool(context, pool)

        LOG.info("Created %(pool)s", {'pool': pool})

        pool = DesignateAdapter.render('API_v2', pool, request=request)
        response.status_int = 201

        response.headers['Location'] = pool['links']['self']

        # Prepare and return the response body
        return pool

    @pecan.expose(template='json:', content_type='application/json')
    @pecan.expose(template='json:', content_type='application/json-patch+json')
    @utils.validate_uuid('pool_id')
    def patch_one(self, pool_id):
        """Update the specific pool"""

        LOG.warning("Use of this API Method is DEPRECATED. This will have "
                    "unforeseen side affects when used with the "
                    "designate-manage pool commands")

        request = pecan.request
        context = request.environ['context']
        body = request.body_dict
        response = pecan.response

        if request.content_type == 'application/json-patch+json':
            raise NotImplementedError('json-patch not implemented')

        # Fetch the existing pool
        pool = self.central_api.get_pool(context, pool_id)

        pool = DesignateAdapter.parse('API_v2', body, pool)

        pool.validate()

        pool = self.central_api.update_pool(context, pool)

        LOG.info("Updated %(pool)s", {'pool': pool})

        response.status_int = 202

        return DesignateAdapter.render('API_v2', pool, request=request)

    @pecan.expose(template=None, content_type='application/json')
    @utils.validate_uuid('pool_id')
    def delete_one(self, pool_id):
        """Delete the specific pool"""

        LOG.warning("Use of this API Method is DEPRECATED. This will have "
                    "unforeseen side affects when used with the "
                    "designate-manage pool commands")

        request = pecan.request
        response = pecan.response
        context = request.environ['context']

        pool = self.central_api.delete_pool(context, pool_id)

        LOG.info("Deleted %(pool)s", {'pool': pool})

        response.status_int = 204

        # NOTE: This is a hack and a half.. But Pecan needs it.
        return ''



# Copyright (C) 2013 eNovance SAS <licensing@enovance.com>
#
# Author: Artom Lifshitz <artom.lifshitz@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from oslo_config import cfg
from oslo_utils import excutils

from designate import backend
from designate.backend import base


LOG = logging.getLogger(__name__)
CFG_GROUP = 'backend:multi'


class MultiBackend(base.Backend):
    """
    Multi-backend backend

    This backend dispatches calls to a master backend and a slave backend.
    It enforces master/slave ordering semantics as follows:

    Creates for tsigkeys, servers and domains are done on the master first,
    then on the slave.

    Updates for tsigkeys, servers and domains and all operations on records
    are done on the master only. It's assumed masters and slaves use an
    external mechanism to sync existing domains, most likely XFR.

    Deletes are done on the slave first, then on the master.

    If the create on the slave fails, the domain/tsigkey/server is deleted from
    the master. If delete on the master fails, the domain/tdigkey/server is
    recreated on the slave.
    """
    __plugin_name__ = 'multi'

    @classmethod
    def get_cfg_opts(cls):
        group = cfg.OptGroup(
            name=CFG_GROUP, title="Configuration for multi-backend Backend"
        )

        opts = [
            cfg.StrOpt('master', default='fake', help='Master backend'),
            cfg.StrOpt('slave', default='fake', help='Slave backend'),
        ]

        return [(group, opts)]

    def __init__(self, central_service):
        super(MultiBackend, self).__init__(central_service)
        self.central = central_service
        self.master = backend.get_backend(cfg.CONF[CFG_GROUP].master,
                                          central_service)
        self.slave = backend.get_backend(cfg.CONF[CFG_GROUP].slave,
                                         central_service)

    def start(self):
        self.master.start()
        self.slave.start()

    def stop(self):
        self.slave.stop()
        self.master.stop()

    def create_tsigkey(self, context, tsigkey):
        self.master.create_tsigkey(context, tsigkey)
        try:
            self.slave.create_tsigkey(context, tsigkey)
        except Exception:
            with excutils.save_and_reraise_exception():
                self.master.delete_tsigkey(context, tsigkey)

    def update_tsigkey(self, context, tsigkey):
        self.master.update_tsigkey(context, tsigkey)

    def delete_tsigkey(self, context, tsigkey):
        self.slave.delete_tsigkey(context, tsigkey)
        try:
            self.master.delete_tsigkey(context, tsigkey)
        except Exception:
            with excutils.save_and_reraise_exception():
                self.slave.create_tsigkey(context, tsigkey)

    def create_zone(self, context, zone):
        self.master.create_zone(context, zone)
        try:
            self.slave.create_zone(context, zone)
        except Exception:
            with excutils.save_and_reraise_exception():
                self.master.delete_zone(context, zone)

    def update_zone(self, context, zone):
        self.master.update_zone(context, zone)

    def delete_zone(self, context, zone):
        # Fetch the full zone from Central first, as we may
        # have to recreate it on slave if delete on master fails
        deleted_context = context.deepcopy()
        deleted_context.show_deleted = True

        full_domain = self.central.find_zone(
            deleted_context, {'id': zone['id']})

        self.slave.delete_zone(context, zone)
        try:
            self.master.delete_zone(context, zone)
        except Exception:
            with excutils.save_and_reraise_exception():
                self.slave.create_zone(context, zone)

                [self.slave.create_record(context, zone, record)
                 for record in self.central.find_records(
                     context, {'domain_id': full_domain['id']})]

    def create_recordset(self, context, zone, recordset):
        self.master.create_recordset(context, zone, recordset)

    def update_recordset(self, context, zone, recordset):
        self.master.update_recordset(context, zone, recordset)

    def delete_recordset(self, context, zone, recordset):
        self.master.delete_recordset(context, zone, recordset)

    def create_record(self, context, zone, recordset, record):
        self.master.create_record(context, zone, recordset, record)

    def update_record(self, context, zone, recordset, record):
        self.master.update_record(context, zone, recordset, record)

    def delete_record(self, context, zone, recordset, record):
        self.master.delete_record(context, zone, recordset, record)

    def ping(self, context):
        return {
            'master': self.master.ping(context),
            'slave': self.slave.ping(context)
        }



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Endre Karlson <endre.karlson@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import socket
from unittest import mock

import dns
import dns.exception
import dns.message
import dns.rcode
import dns.rdatatype
import dns.zone
import eventlet
from oslo_config import cfg
from oslo_config import fixture as cfg_fixture
import oslotest.base

from designate import dnsutils
from designate import exceptions

CONF = cfg.CONF


class TestDNSUtils(oslotest.base.BaseTestCase):
    def setUp(self):
        super(TestDNSUtils, self).setUp()
        self.useFixture(cfg_fixture.Config(CONF))

    @mock.patch.object(dns.query, 'udp')
    def test_send_udp_dns_message(self, mock_udp):
        CONF.set_override('all_tcp', False, 'service:mdns')
        dnsutils.send_dns_message('msg', '192.0.2.1', 1234, 1)
        mock_udp.assert_called_with(
            'msg', '192.0.2.1', port=1234, timeout=1
        )

    @mock.patch.object(dns.query, 'tcp')
    def test_send_tcp_dns_message(self, mock_tcp):
        CONF.set_override('all_tcp', True, 'service:mdns')
        dnsutils.send_dns_message('msg', '192.0.2.1', 1234, 1)
        mock_tcp.assert_called_with(
            'msg', '192.0.2.1', port=1234, timeout=1
        )

    def test_all_tcp_default(self):
        self.assertEqual(False, dnsutils.use_all_tcp())

    def test_all_tcp_using_mdns(self):
        CONF.set_override('all_tcp', True, 'service:mdns')
        self.assertEqual(True, dnsutils.use_all_tcp())

    def test_all_tcp_using_worker(self):
        CONF.set_override('all_tcp', True, 'service:worker')
        self.assertEqual(True, dnsutils.use_all_tcp())

    def test_xfr_default(self):
        self.assertEqual(10, dnsutils.xfr_timeout())

    def test_xfr_timeout_set_using_mdns(self):
        CONF.set_override('xfr_timeout', 30, 'service:mdns')
        self.assertEqual(30, dnsutils.xfr_timeout())

    def test_xfr_timeout_set_using_worker(self):
        CONF.set_override('xfr_timeout', 40, 'service:worker')
        self.assertEqual(40, dnsutils.xfr_timeout())


class TestDoAfxr(oslotest.base.BaseTestCase):
    def setUp(self):
        super(TestDoAfxr, self).setUp()
        self.useFixture(cfg_fixture.Config(CONF))

    @mock.patch.object(dns.query, 'xfr')
    @mock.patch.object(dns.zone, 'from_xfr')
    def test_do_afxr(self, mock_from_xfr_impl, mock_xfr):
        mock_from_xfr = mock.MagicMock()
        mock_from_xfr_impl.return_value = mock_from_xfr

        mock_from_xfr.origin.to_text.return_value = 'raw_zone'
        mock_from_xfr.return_value = 'raw_zone'

        masters = [
            {'host': '192.168.0.1', 'port': 53},
            {'host': '192.168.0.2', 'port': 53},
        ]

        self.assertEqual(
            mock_from_xfr,
            dnsutils.do_axfr('example.com', masters)
        )

        self.assertTrue(mock_xfr.called)
        self.assertTrue(mock_from_xfr_impl.called)

    def test_do_afxr_no_masters(self):
        masters = [
        ]

        self.assertRaisesRegex(
            exceptions.XFRFailure,
            r'XFR failed for example.com. No servers in \[\] was reached.',
            dnsutils.do_axfr, 'example.com', masters,
        )

    @mock.patch.object(dns.query, 'xfr')
    @mock.patch.object(dns.zone, 'from_xfr')
    @mock.patch.object(eventlet.Timeout, 'cancel')
    def test_do_afxr_fails_with_timeout(self, mock_cancel, mock_from_xfr,
                                        mock_xfr):
        mock_from_xfr.side_effect = eventlet.Timeout()

        masters = [
            {'host': '192.168.0.1', 'port': 53},
            {'host': '192.168.0.2', 'port': 53},
            {'host': '192.168.0.3', 'port': 53},
            {'host': '192.168.0.4', 'port': 53},
        ]

        self.assertRaises(
            exceptions.XFRFailure,
            dnsutils.do_axfr, 'example.com.', masters,
        )

        self.assertTrue(mock_xfr.called)
        self.assertTrue(mock_from_xfr.called)
        self.assertTrue(mock_cancel.called)

    @mock.patch.object(dns.query, 'xfr')
    @mock.patch.object(dns.zone, 'from_xfr')
    def test_do_afxr_fails_with_form_error(self, mock_from_xfr, mock_xfr):
        mock_from_xfr.side_effect = dns.exception.FormError()

        masters = [
            {'host': '192.168.0.1', 'port': 53},
        ]

        self.assertRaises(
            exceptions.XFRFailure,
            dnsutils.do_axfr, 'example.com.', masters,
        )

        self.assertTrue(mock_xfr.called)
        self.assertTrue(mock_from_xfr.called)

    @mock.patch.object(dns.query, 'xfr')
    @mock.patch.object(dns.zone, 'from_xfr')
    def test_do_afxr_fails_with_socket_error(self, mock_from_xfr, mock_xfr):
        mock_from_xfr.side_effect = socket.error()

        masters = [
            {'host': '192.168.0.1', 'port': 53},
        ]

        self.assertRaises(
            exceptions.XFRFailure,
            dnsutils.do_axfr, 'example.com.', masters,
        )

        self.assertTrue(mock_xfr.called)
        self.assertTrue(mock_from_xfr.called)

    @mock.patch.object(dns.query, 'xfr')
    @mock.patch.object(dns.zone, 'from_xfr')
    def test_do_afxr_fails_with_exception(self, mock_from_xfr, mock_xfr):
        mock_from_xfr.side_effect = Exception()

        masters = [
            {'host': '192.168.0.1', 'port': 53},
        ]

        self.assertRaises(
            exceptions.XFRFailure,
            dnsutils.do_axfr, 'example.com.', masters,
        )

        self.assertTrue(mock_xfr.called)
        self.assertTrue(mock_from_xfr.called)


class TestDNSMessages(oslotest.base.BaseTestCase):
    def setUp(self):
        super(TestDNSMessages, self).setUp()
        self.useFixture(cfg_fixture.Config(CONF))

    @mock.patch.object(dnsutils, 'send_dns_message')
    def test_notify(self, mock_send_dns_message):
        dnsutils.notify('notify.test.', 'localhost', port=54)

        mock_send_dns_message.assert_called_with(
            mock.ANY, 'localhost', port=54, timeout=10
        )

        query = mock_send_dns_message.call_args[0][0]
        txt = query.to_text().split('\n')[1:]
        self.assertEqual([
            'opcode NOTIFY',
            'rcode NOERROR',
            'flags RD',
            ';QUESTION',
            'notify.test. IN SOA',
            ';ANSWER',
            ';AUTHORITY',
            ';ADDITIONAL'
        ], txt)

    @mock.patch.object(dnsutils, 'send_dns_message')
    def test_soa(self, mock_send_dns_message):
        dnsutils.soa_query('soa.test.', 'localhost', port=54)

        mock_send_dns_message.assert_called_with(
            mock.ANY, 'localhost', port=54, timeout=10
        )

        query = mock_send_dns_message.call_args[0][0]
        txt = query.to_text().split('\n')[1:]
        self.assertEqual([
            'opcode QUERY',
            'rcode NOERROR',
            'flags RD',
            ';QUESTION',
            'soa.test. IN SOA',
            ';ANSWER',
            ';AUTHORITY',
            ';ADDITIONAL'
        ], txt)

    @mock.patch.object(dnsutils, 'send_dns_message')
    def test_get_serial(self, mock_send_dns_message):
        mock_rdataset = mock.Mock(serial=5)
        mock_answer = mock.Mock()
        mock_answer.to_rdataset.return_value = [mock_rdataset]

        mock_result = mock.Mock()
        mock_result.answer = [mock_answer]
        mock_send_dns_message.return_value = mock_result

        self.assertEqual(
            5, dnsutils.get_serial('serial.test.', 'localhost', port=54)
        )

        query = mock_send_dns_message.call_args[0][0]
        txt = query.to_text().split('\n')[1:]
        self.assertEqual([
            'opcode QUERY',
            'rcode NOERROR',
            'flags RD',
            ';QUESTION',
            'serial.test. IN SOA',
            ';ANSWER',
            ';AUTHORITY',
            ';ADDITIONAL'
        ], txt)

    @mock.patch.object(dnsutils, 'send_dns_message')
    def test_get_serial_no_answer(self, mock_send_dns_message):
        mock_result = mock.Mock()
        mock_result.answer = []
        mock_send_dns_message.return_value = mock_result

        self.assertFalse(
            dnsutils.get_serial('serial.test.', 'localhost', port=54)
        )

    @mock.patch.object(dnsutils, 'send_dns_message')
    def test_get_serial_no_rdataset(self, mock_send_dns_message):
        mock_answer = mock.Mock()
        mock_answer.to_rdataset.return_value = []

        mock_result = mock.Mock()
        mock_result.answer = [mock_answer]
        mock_send_dns_message.return_value = mock_result

        self.assertFalse(
            dnsutils.get_serial('serial.test.', 'localhost', port=54)
        )



#!/usr/bin/env python
# Copyright (C) 2018 Verizon
#
# Author: Graham Hayes <gr@ham.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import argparse
import ipaddress
import logging
import sys

import dns.exception
from dns.ipv4 import inet_aton
from keystoneauth1.identity import generic
from keystoneauth1 import session as keystone_session

from designateclient import shell
from designateclient.v2 import client


auth = generic.Password(
    auth_url=shell.env('OS_AUTH_URL'),
    username=shell.env('OS_USERNAME'),
    password=shell.env('OS_PASSWORD'),
    project_name=shell.env('OS_PROJECT_NAME'),
    project_domain_id=shell.env('OS_PROJECT_DOMAIN_ID'),
    user_domain_id=shell.env('OS_USER_DOMAIN_ID'))

session = keystone_session.Session(auth=auth)

client = client.Client(session=session)

logging.basicConfig()
LOG = logging.getLogger('fixleadingzeros')


def find_bad_recordsets():
    bad_recordsets = {}
    LOG.debug("Looking for all A recordsets")
    recordsets = client.recordsets.list_all_zones(criterion={'type': 'A', })
    LOG.debug("Found %d A recordsets", len(recordsets))
    LOG.debug("Filtering recordsets")
    for recordset in recordsets:
        for record in recordset['records']:
            try:
                inet_aton(record)
            except dns.exception.SyntaxError:
                bad_recordsets[recordset['id']] = recordset
    LOG.debug("Found %d A invaild recordsets", len(bad_recordsets))
    return bad_recordsets


def show_recordsets(recordsets):
    for rs in recordsets:
        LOG.info(
            ("%(name)s - %(records)s - Zone ID: %(zone_id)s - "
             "Project ID: %(project_id)s ") % recordsets[rs])


def fix_bad_recordsets(bad_recordsets):
    LOG.debug("Removing leading zeros in IPv4 addresses")
    for rs in bad_recordsets:
        new_records = []
        for ip in bad_recordsets[rs]['records']:
            ip = '.'.join(f'{int(i)}' for i in ip.split('.'))
            new_records.append(
                str(ipaddress.IPv4Address(ip))
            )
        bad_recordsets[rs]['records'] = new_records
    return bad_recordsets


def update_recordsets(recordsets):
    LOG.info("Updating recordsets")
    for rs in recordsets:
        LOG.debug(("Updating %(name)s - %(records)s - Zone ID: %(zone_id)s - "
                   "Project ID: %(project_id)s ") % recordsets[rs])
        client.recordsets.update(
            recordsets[rs]['zone_id'],
            recordsets[rs]['id'],
            {'records': recordsets[rs]['records']}
        )


def main():
    parser = argparse.ArgumentParser(
        description='Fix any recordsets that have leading zeros in A records')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='verbose output')
    parser.add_argument('-d', '--dry-run', action='store_true',
                        help='do not modify records, just log bad records')
    parser.add_argument('-a', '--all-projects', action='store_true',
                        help="Run on all projects")
    args = parser.parse_args()
    if args.verbose:
        LOG.setLevel(logging.DEBUG)
    else:
        LOG.setLevel(logging.INFO)

    if args.all_projects:
        client.session.all_projects = True

    bad_recordsets = find_bad_recordsets()

    LOG.info("Bad recordsets")
    show_recordsets(bad_recordsets)

    fixed_recordsets = fix_bad_recordsets(bad_recordsets)
    LOG.info("Fixed recordsets")
    show_recordsets(fixed_recordsets)

    if not args.dry_run:
        update_recordsets(fixed_recordsets)


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging

from designate.backend import base

LOG = logging.getLogger(__name__)

GOOD_STATUSES = [
    'integrated', 'grades.master-compatible', 'release-compatible'
]


def get_backend(target):
    cls = base.Backend.get_driver(target.type)

    message = "Backend Driver '%s' loaded. Has status of '%s'" % (
        target.type, cls.__backend_status__
    )

    if cls.__backend_status__ in GOOD_STATUSES:
        LOG.info(message)
    else:
        LOG.warning(message)

    return cls(target)



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Graham Hayes <graham.hayes@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from designate.tests.test_api.test_v2 import ApiV2TestCase


class ApiV2ZoneTransfersTest(ApiV2TestCase):
    def setUp(self):
        super(ApiV2ZoneTransfersTest, self).setUp()

        self.zone = self.create_zone()
        self.tenant_1_context = self.get_context(project_id=1)
        self.tenant_2_context = self.get_context(project_id=2)
        self.policy({'admin': '@'})

    def test_create_zone_transfer_request(self):
        response = self.client.post_json(
            '/zones/%s/tasks/transfer_requests' % self.zone.id,
            {})

        # Check the headers are what we expect
        self.assertEqual(201, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIn('created_at', response.json)
        self.assertEqual('ACTIVE', response.json['status'])
        self.assertEqual(
            self.zone.name,
            response.json['zone_name'])
        self.assertEqual(
            self.zone.id,
            response.json['zone_id'])
        self.assertIsNone(response.json['updated_at'])

    def test_create_zone_transfer_request_scoped(self):
        response = self.client.post_json(
            '/zones/%s/tasks/transfer_requests' % self.zone.id,
            {'target_project_id': str(self.tenant_1_context.project_id)})

        # Check the headers are what we expect
        self.assertEqual(201, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIn('created_at', response.json)
        self.assertEqual('ACTIVE', response.json['status'])
        self.assertEqual(
            str(self.tenant_1_context.project_id),
            response.json['target_project_id'])
        self.assertEqual(
            self.zone.id,
            response.json['zone_id'])
        self.assertIsNone(response.json['updated_at'])

    def test_create_zone_transfer_request_empty_body(self):
        # Send an empty ("None") body
        response = self.client.post_json(
            '/zones/%s/tasks/transfer_requests' % self.zone.id,
            None)

        # Check the headers are what we expect
        self.assertEqual(201, response.status_int)
        self.assertEqual('application/json', response.content_type)

    def test_get_zone_transfer_request(self):
        initial = self.client.post_json(
            '/zones/%s/tasks/transfer_requests' % self.zone.id,
            {})

        response = self.client.get(
            '/zones/tasks/transfer_requests/%s' %
            (initial.json['id']))

        # Check the headers are what we expect
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIn('created_at', response.json)
        self.assertEqual('ACTIVE', response.json['status'])
        self.assertEqual(
            self.zone.id,
            response.json['zone_id'])
        self.assertIn('updated_at', response.json)

    def test_get_zone_transfer_requests(self):
        response = self.client.get(
            '/zones/tasks/transfer_requests')

        # Check the headers are what we expect
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('transfer_requests', response.json)
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # We should start with 0 transfer accepts
        self.assertEqual(0, len(response.json['transfer_requests']))

        self.client.post_json(
            '/zones/%s/tasks/transfer_requests' % self.zone.id,
            {})

        data = self.client.get(
            '/zones/tasks/transfer_requests')

        self.assertEqual(1, len(data.json['transfer_requests']))

    def test_update_zone_transfer_request(self):
        initial = self.client.post_json(
            '/zones/%s/tasks/transfer_requests' % self.zone.id,
            {})

        response = self.client.patch_json(
            '/zones/tasks/transfer_requests/%s' %
            (initial.json['id']),
            {"description": "TEST"})

        # Check the headers are what we expect
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIn('created_at', response.json)
        self.assertEqual('ACTIVE', response.json['status'])
        self.assertEqual(
            self.zone.id,
            response.json['zone_id'])
        self.assertEqual(
            'TEST', response.json['description'])
        self.assertIn('updated_at', response.json)

    def test_delete_zone_transfer_request(self):
        initial = self.client.post_json(
            '/zones/%s/tasks/transfer_requests' % self.zone.id,
            {})

        response = self.client.delete(
            '/zones/tasks/transfer_requests/%s' %
            (initial.json['id']))

        # Check the headers are what we expect
        self.assertEqual(204, response.status_int)

    def test_create_zone_transfer_accept(self):
        initial = self.client.post_json(
            '/zones/%s/tasks/transfer_requests' % self.zone.id,
            {})

        response = self.client.post_json(
            '/zones/tasks/transfer_accepts',
            {
                'zone_transfer_request_id':
                    initial.json['id'],
                'key': initial.json['key']
            })

        new_ztr = self.client.get(
            '/zones/tasks/transfer_requests/%s' %
            (initial.json['id']))

        # Check the headers are what we expect
        self.assertEqual(201, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])
        self.assertIn('zone', response.json['links'])

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertEqual(
            'COMPLETE',
            response.json['status'])

        self.assertEqual(
            'COMPLETE',
            new_ztr.json['status'])

    def test_get_zone_transfer_accept(self):
        initial = self.client.post_json(
            '/zones/%s/tasks/transfer_requests' % self.zone.id,
            {})

        transfer_accept = self.client.post_json(
            '/zones/tasks/transfer_accepts',
            {
                'zone_transfer_request_id':
                    initial.json['id'],
                'key': initial.json['key']
            })
        response = self.client.get(
            '/zones/tasks/transfer_accepts/%s' %
            (transfer_accept.json['id']))

        # Check the headers are what we expect
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIn('created_at', response.json)
        self.assertEqual('COMPLETE', response.json['status'])
        self.assertEqual(
            self.zone.id,
            response.json['zone_id'])
        self.assertIn('updated_at', response.json)
        self.assertIn('key', response.json)
        self.assertIn(initial.json['id'],
                      response.json['zone_transfer_request_id'])

    def test_get_zone_transfer_accept_invalid_id(self):
        self._assert_invalid_uuid(self.client.get,
            '/zones/tasks/transfer_accepts/%s')

    def test_get_zone_transfer_accepts(self):
        response = self.client.get(
                       '/zones/tasks/transfer_accepts')

        # Check the headers are what we expect
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('transfer_accepts', response.json)
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])
        self.assertIn('metadata', response.json)

        # We should start with 0 transfer accepts
        self.assertEqual(0, len(response.json['transfer_accepts']))

        initial = self.client.post_json(
            '/zones/%s/tasks/transfer_requests' % self.zone.id,
            {})

        self.client.post_json(
            '/zones/tasks/transfer_accepts',
            {
                'zone_transfer_request_id':
                    initial.json['id'],
                'key': initial.json['key']
            })

        data = self.client.get(
                       '/zones/tasks/transfer_accepts')
        self.assertEqual(1, len(data.json['transfer_accepts']))

    def test_create_zone_transfer_request_deleting_zone(self):
        url = '/zones/%s/tasks/transfer_requests' % self.zone.id
        body = {}
        self.client.delete('/zones/%s' % self.zone['id'], status=202)
        self._assert_exception('bad_request', 400, self.client.post_json, url,
                               body)

    def test_create_zone_transfer_accept_deleting_zone(self):
        url = '/zones/%s/tasks/transfer_requests' % self.zone.id
        body = {}
        self.client.delete('/zones/%s' % self.zone['id'], status=202)
        self._assert_exception('bad_request', 400, self.client.post_json, url,
                               body)

    # Metadata tests
    def test_metadata_exists_zone_transfer_accepts(self):
        initial = self.client.post_json(
            '/zones/%s/tasks/transfer_requests' % self.zone.id,
            {})

        self.client.post_json(
            '/zones/tasks/transfer_accepts',
            {
                'zone_transfer_request_id':
                    initial.json['id'],
                'key': initial.json['key']
            })

        result = self.client.get(
            '/zones/tasks/transfer_accepts')

        # Make sure the fields exist
        self.assertIn('metadata', result.json)
        self.assertIn('total_count', result.json['metadata'])

    def test_total_count_zone_transfer_accepts(self):
        initial = self.client.post_json(
            '/zones/%s/tasks/transfer_requests' % self.zone.id,
            {})

        self.client.post_json(
            '/zones/tasks/transfer_accepts',
            {
                'zone_transfer_request_id':
                    initial.json['id'],
                'key': initial.json['key']
            })

        result = self.client.get(
            '/zones/tasks/transfer_accepts')

        # Make sure total_count picked it up
        self.assertEqual(1, result.json['metadata']['total_count'])



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging

from designate.common import constants
from designate import quota
from designate import tests

LOG = logging.getLogger(__name__)


class StorageQuotaTest(tests.TestCase):
    def setUp(self):
        super(StorageQuotaTest, self).setUp()
        self.config(quota_driver='storage')
        self.quota = quota.get_quota()

    def test_set_quota_create_min(self):
        context = self.get_admin_context()
        context.all_tenants = True

        for current_quota in constants.VALID_QUOTAS:
            quota = self.quota.set_quota(context, 'tenant_id',
                                         current_quota, constants.MIN_QUOTA)

            self.assertEqual({current_quota: constants.MIN_QUOTA}, quota)

            # Drop into the storage layer directly to ensure the quota was
            # created successfully
            criterion = {
                'tenant_id': 'tenant_id',
                'resource': current_quota
            }

            quota = self.quota.storage.find_quota(context, criterion)

            self.assertEqual('tenant_id', quota['tenant_id'])
            self.assertEqual(current_quota, quota['resource'])
            self.assertEqual(constants.MIN_QUOTA, quota['hard_limit'])

    def test_set_quota_create_max(self):
        context = self.get_admin_context()
        context.all_tenants = True

        for current_quota in constants.VALID_QUOTAS:
            quota = self.quota.set_quota(context, 'tenant_id',
                                         current_quota, constants.MAX_QUOTA)

            self.assertEqual({current_quota: constants.MAX_QUOTA}, quota)

            # Drop into the storage layer directly to ensure the quota was
            # created successfully
            criterion = {
                'tenant_id': 'tenant_id',
                'resource': current_quota
            }

            quota = self.quota.storage.find_quota(context, criterion)

            self.assertEqual('tenant_id', quota['tenant_id'])
            self.assertEqual(current_quota, quota['resource'])
            self.assertEqual(constants.MAX_QUOTA, quota['hard_limit'])

    def test_set_quota_update(self):
        context = self.get_admin_context()
        context.all_tenants = True

        for current_quota in constants.VALID_QUOTAS:
            # First up, Create the quota
            self.quota.set_quota(context, 'tenant_id', current_quota, 1500)

            # Next, update the quota
            self.quota.set_quota(context, 'tenant_id', current_quota, 1234)

            # Drop into the storage layer directly to ensure the quota was
            # updated successfully
            criterion = {
                'tenant_id': 'tenant_id',
                'resource': current_quota
            }

            quota = self.quota.storage.find_quota(context, criterion)

            self.assertEqual('tenant_id', quota['tenant_id'])
            self.assertEqual(current_quota, quota['resource'])
            self.assertEqual(1234, quota['hard_limit'])

    def test_reset_quotas(self):
        context = self.get_admin_context()
        context.all_tenants = True

        # First up, Create a zones quota
        self.quota.set_quota(context, 'tenant_id', 'zones', 1500)

        # Then, Create a zone_records quota
        self.quota.set_quota(context, 'tenant_id', 'zone_records', 800)

        # Now, Reset the tenants quota
        self.quota.reset_quotas(context, 'tenant_id')

        # Drop into the storage layer directly to ensure the tenant has no
        # specific quotas registed.
        criterion = {
            'tenant_id': 'tenant_id'
        }

        quotas = self.quota.storage.find_quotas(context, criterion)
        self.assertEqual(0, len(quotas))



# Copyright 2016 Hewlett Packard Enterprise Development Company LP
#
# Author: Federico Ceratto <federico.ceratto@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
backend.agent_backend.impl_gdnsd
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
gdnsd agent backend

Create, update, delete zones locally on a gdnsd resolver using the
gdnsd utility.

Supported Knot versions: >= 2.1, < 3

`User documentation <../../admin/backends/gdnsd_agent.html>`_

.. WARNING::

    Untested, do not use in production.

.. NOTE::

    If the backend is killed during a configuration transaction it might be
    required to manually abort the transaction with `sudo gdnsd conf-abort`

Configured in [service:agent:gdnsd]
"""

import errno
import os
import string
import tempfile

import dns
import dns.resolver
from oslo_concurrency.processutils import ProcessExecutionError
from oslo_config import cfg
from oslo_log import log as logging

from designate.backend.agent_backend import base
from designate import exceptions
from designate import utils

CFG_GROUP_NAME = 'backend:agent:gdnsd'
LOG = logging.getLogger(__name__)
# rootwrap requires a command name instead of full path
GDNSD_DEFAULT_PATH = 'gdnsd'
CONFDIR_PATH = '/etc/gdnsd'
SOA_QUERY_TIMEOUT = 1
ZONE_FILE_PERMISSIONS = 0o0644


def filter_exceptions(fn):
    # Let Backend() exceptions pass through, log out every other exception
    # and re-raise it as Backend()
    def wrapper(*a, **kw):
        try:
            return fn(*a, **kw)
        except exceptions.Backend as e:
            raise e
        except Exception as e:
            LOG.error("Unhandled exception %s", e, exc_info=True)
            raise exceptions.Backend(e)

    return wrapper


class GdnsdBackend(base.AgentBackend):
    __plugin_name__ = 'gdnsd'
    __backend_status__ = 'experimental'

    def __init__(self, *a, **kw):
        """Configure the backend"""
        super(GdnsdBackend, self).__init__(*a, **kw)

        self._gdnsd_cmd_name = cfg.CONF[CFG_GROUP_NAME].gdnsd_cmd_name
        LOG.info("gdnsd command: %r", self._gdnsd_cmd_name)
        self._confdir_path = cfg.CONF[CFG_GROUP_NAME].confdir_path
        self._zonedir_path = os.path.join(self._confdir_path, 'zones')
        LOG.info("gdnsd conf directory: %r", self._confdir_path)
        self._resolver = dns.resolver.Resolver(configure=False)
        self._resolver.timeout = SOA_QUERY_TIMEOUT
        self._resolver.lifetime = SOA_QUERY_TIMEOUT
        self._resolver.nameservers = [
            cfg.CONF[CFG_GROUP_NAME].query_destination
        ]
        LOG.info("Resolvers: %r", self._resolver.nameservers)
        self._check_dirs(self._zonedir_path)

    def start(self):
        """Start the backend, check gdnsd configuration

        :raises: exception.Backend on invalid configuration
        """
        LOG.info("Started gdnsd backend")
        self._check_conf()

    def _check_conf(self):
        """Run gdnsd to check its configuration
        """
        try:
            out, err = utils.execute(
                cfg.CONF[CFG_GROUP_NAME].gdnsd_cmd_name,
                '-D', '-x', 'checkconf', '-c', self._confdir_path,
                run_as_root=False,
            )
        except ProcessExecutionError as e:
            LOG.error("Command output: %(out)r Stderr: %(err)r",
                      {
                          'out': e.stdout,
                          'err': e.stderr
                      })
            raise exceptions.Backend("Configuration check failed")

    def _check_dirs(self, *dirnames):
        """Check if directories are writable
        """
        for dn in dirnames:
            if not os.path.isdir(dn):
                raise exceptions.Backend("Missing directory %s" % dn)
            if not os.access(dn, os.W_OK):
                raise exceptions.Backend("Directory not writable: %s" % dn)

    def find_zone_serial(self, zone_name):
        """Query the local resolver for a zone
        Times out after SOA_QUERY_TIMEOUT
        """
        LOG.debug("Finding %s", zone_name)
        try:
            rdata = self._resolver.query(
                zone_name, rdtype=dns.rdatatype.SOA)[0]
            return rdata.serial
        except Exception:
            return None

    def _generate_zone_filename(self, zone_name):
        """Generate a filename for a zone file
        "/" is traslated into "@"
        Non-valid characters are translated into \\ NNN
        where NNN is a decimal integer in the range 0 - 255
        The filename is lowercase

        :returns: valid filename (string)
        """
        valid_chars = "-_.@%s%s" % (string.ascii_letters, string.digits)
        fname = zone_name.replace('/', '@').lower()
        fname = [c if c in valid_chars else "\03%d" % ord(c)
                 for c in fname]
        return ''.join(fname)

    def _write_zone_file(self, zone):
        """Create or update a zone file atomically.
        The zone file is written to a unique temp file and then renamed
        """
        zone_name = zone.origin.to_text(omit_final_dot=True)
        if isinstance(zone_name, bytes):
            zone_name = zone_name.decode('utf-8')
        zone_base_fname = self._generate_zone_filename(zone_name)
        zone_fname = os.path.join(self._zonedir_path, zone_base_fname)
        try:
            # gdnsd ignores hidden files
            tmp_zone_fname = tempfile.mkstemp(
                prefix=".%s" % zone_base_fname,
                dir=self._zonedir_path,
            )[1]
            LOG.debug("Writing zone %r to %r and renaming it to %r",
                      zone_name, tmp_zone_fname, zone_fname)
            zone.to_file(tmp_zone_fname)
            os.chmod(tmp_zone_fname, ZONE_FILE_PERMISSIONS)
            os.rename(tmp_zone_fname, zone_fname)
        finally:
            try:
                os.remove(tmp_zone_fname)
            except OSError:
                pass

    @filter_exceptions
    def create_zone(self, zone):
        """Create a new Zone
        Do not raise exceptions if the zone already exists.

        :param zone: zone to be  created
        :type zone: raw pythondns Zone
        :raises: exceptions.Backend on error
        """
        # The zone might be already in place due to a race condition between
        # checking if the zone is there and creating it across different
        # greenlets
        self._write_zone_file(zone)

    @filter_exceptions
    def update_zone(self, zone):
        """Instruct Djbdns DNS to perform AXFR from MiniDNS

        :param zone: zone to be  created
        :type zone: raw pythondns Zone
        :raises: exceptions.Backend on error
        """
        self._write_zone_file(zone)

    @filter_exceptions
    def delete_zone(self, zone_name):
        """Delete a new Zone
        Do not raise exceptions if the zone does not exist.

        :param zone_name: zone name
        :type zone_name: str
        :raises: exceptions.Backend on error
        """
        zone_name = zone_name.rstrip('.')
        LOG.debug('Deleting Zone: %s', zone_name)
        zone_fn = self._generate_zone_filename(zone_name)
        zone_fn = os.path.join(self._zonedir_path, zone_fn)
        try:
            os.remove(zone_fn)
            LOG.debug('Deleted Zone: %s', zone_name)
        except OSError as e:
            if errno.ENOENT == e.errno:
                LOG.info("Zone datafile %s was already deleted", zone_fn)
                return
            raise



# Copyright 2021 NS1 Inc. https://www.ns1.com
#
# Author: Dragan Blagojevic <dblagojevic@daitan.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import oslotest.base
import requests_mock

from designate.backend import impl_ns1
from designate import context
from designate import exceptions
from designate import objects
from designate.tests import fixtures


class NS1BackendTestCase(oslotest.base.BaseTestCase):
    def setUp(self):
        super(NS1BackendTestCase, self).setUp()
        self.stdlog = fixtures.StandardLogging()
        self.useFixture(self.stdlog)

        self.context = mock.Mock()
        self.admin_context = mock.Mock()
        mock.patch.object(
            context.DesignateContext, 'get_admin_context',
            return_value=self.admin_context).start()

        self.api_address = 'https://192.0.2.3/v1/zones/example.com'
        self.zone = objects.Zone(
            id='e2bed4dc-9d01-11e4-89d3-123b93f75cba',
            name='example.com.',
            email='example@example.com',
        )
        self.target = {
            'id': '4588652b-50e7-46b9-b688-a9bad40a873e',
            'type': 'ns1',
            'masters': [
                {'host': '192.0.2.1', 'port': 53},
                {'host': '192.0.2.2', 'port': 35},
            ],
            'options': [
                {'key': 'api_endpoint', 'value': '192.0.2.3'},
                {'key': 'api_token', 'value': 'test_key'},
            ],
        }
        self.target_tsig = {
            'id': '4588652b-50e7-46b9-b688-a9bad40a873e',
            'type': 'ns1',
            'masters': [
                {'host': '192.0.2.1', 'port': 53},
                {'host': '192.0.2.2', 'port': 35},
            ],
            'options': [
                {'key': 'api_endpoint', 'value': '192.0.2.3'},
                {'key': 'api_token', 'value': 'test_key'},
                {'key': 'tsigkey_name', 'value': 'test_key'},
                {'key': 'tsigkey_hash', 'value': 'hmac-sha512'},
                {'key': 'tsigkey_value', 'value': 'aaaabbbbccc'},
            ],
        }
        self.put_request_json = {
            'zone': u'example.com',
            'secondary': {
                'enabled': True,
                'primary_ip': '192.0.2.1',
                'primary_port': 53
            }
        }
        self.put_request_tsig_json = {
            'zone': u'example.com',
            'secondary': {
                'enabled': True,
                'primary_ip': '192.0.2.1',
                'primary_port': 53,
                'tsig': {
                    'enabled': True,
                    'hash': 'hmac-sha512',
                    'name': 'test_key',
                    'key': 'aaaabbbbccc'
                }
            }
        }

        self.backend = impl_ns1.NS1Backend(
            objects.PoolTarget.from_dict(self.target)
        )
        self.backend_tsig = impl_ns1.NS1Backend(
            objects.PoolTarget.from_dict(self.target_tsig)
        )

    @requests_mock.mock()
    def test_create_zone_success(self, req_mock):
        req_mock.put(self.api_address)
        req_mock.get(
            self.api_address,
            status_code=404
        )

        self.backend.create_zone(self.context, self.zone)

        self.assertEqual(
            req_mock.last_request.json(),
            self.put_request_json
        )

        self.assertEqual(
            req_mock.last_request.headers.get('X-NSONE-Key'), 'test_key'
        )

    @requests_mock.mock()
    def test_create_zone_with_tsig_success(self, req_mock):
        req_mock.put(self.api_address)
        req_mock.get(
            self.api_address,
            status_code=404
        )

        self.backend_tsig.create_zone(self.context, self.zone)

        self.assertEqual(
            req_mock.last_request.json(),
            self.put_request_tsig_json
        )

        self.assertEqual(
            req_mock.last_request.headers.get('X-NSONE-Key'), 'test_key'
        )

    @requests_mock.mock()
    def test_create_zone_already_exists(self, req_mock):

        req_mock.get(self.api_address, status_code=200)
        req_mock.put(self.api_address)

        self.backend.create_zone(self.context, self.zone)

        self.assertIn(
            "Can't create zone example.com. because it already exists",
            self.stdlog.logger.output
        )

        self.assertEqual(
            req_mock.last_request.headers.get('X-NSONE-Key'), 'test_key'
        )

    @requests_mock.mock()
    def test_create_zone_fail(self, req_mock):
        req_mock.put(
            self.api_address,
            status_code=500,
        )
        req_mock.get(
            self.api_address,
            status_code=404,
        )

        self.assertRaisesRegex(
            exceptions.Backend,
            '500 Server Error: None for url: '
            '%s' % self.api_address,
            self.backend.create_zone, self.context, self.zone
        )

        self.assertEqual(
            req_mock.last_request.headers.get('X-NSONE-Key'), 'test_key'
        )

    @requests_mock.mock()
    def test_delete_zone_success(self, req_mock):
        req_mock.delete(self.api_address, status_code=200)
        req_mock.get(self.api_address, status_code=200)

        self.backend.delete_zone(self.context, self.zone)

        self.assertEqual(
            req_mock.last_request.headers.get('X-NSONE-Key'), 'test_key'
        )

    @requests_mock.mock()
    def test_delete_zone_missing(self, req_mock):
        req_mock.delete(self.api_address, status_code=200)
        req_mock.get(self.api_address, status_code=404)

        self.backend.delete_zone(self.context, self.zone)

        self.assertIn(
            "Trying to delete zone "
            "<Zone id:'e2bed4dc-9d01-11e4-89d3-123b93f75cba' type:'None' "
            "name:'example.com.' pool_id:'None' serial:'None' action:'None' "
            "status:'None'> "
            "but that zone is not "
            "present in the ns1 backend. Assuming success.",
            self.stdlog.logger.output
        )

        self.assertEqual(
            req_mock.last_request.headers.get('X-NSONE-Key'), 'test_key'
        )

    @requests_mock.mock()
    def test_delete_zone_fail(self, req_mock):
        req_mock.delete(self.api_address, status_code=500)
        req_mock.get(self.api_address, status_code=200)

        self.assertRaisesRegex(
            exceptions.Backend,
            '500 Server Error: None for url: '
            '%s' % self.api_address,
            self.backend.delete_zone, self.context, self.zone
        )
        self.assertEqual(
            req_mock.last_request.headers.get('X-NSONE-Key'), 'test_key'
        )



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_log import versionutils
from oslo_policy import policy

from designate.common.policies import base


deprecated_all_tenants = policy.DeprecatedRule(
    name="all_tenants",
    check_str=base.RULE_ADMIN,
    deprecated_reason=base.DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_edit_managed_records = policy.DeprecatedRule(
    name="edit_managed_records",
    check_str=base.RULE_ADMIN,
    deprecated_reason=base.DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_use_low_ttl = policy.DeprecatedRule(
    name="use_low_ttl",
    check_str=base.RULE_ADMIN,
    deprecated_reason=base.DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_use_sudo = policy.DeprecatedRule(
    name="use_sudo",
    check_str=base.RULE_ADMIN,
    deprecated_reason=base.DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)

rules = [
    policy.RuleDefault(
        name="all_tenants",
        check_str=base.SYSTEM_ADMIN,
        scope_types=['system'],
        description='Action on all tenants.',
        deprecated_rule=deprecated_all_tenants),
    policy.RuleDefault(
        name="edit_managed_records",
        check_str=base.SYSTEM_ADMIN,
        scope_types=['system'],
        description='Edit managed records.',
        deprecated_rule=deprecated_edit_managed_records),
    policy.RuleDefault(
        name="use_low_ttl",
        check_str=base.SYSTEM_ADMIN,
        scope_types=['system'],
        description='Use low TTL.',
        deprecated_rule=deprecated_use_low_ttl),
    policy.RuleDefault(
        name="use_sudo",
        check_str=base.SYSTEM_ADMIN,
        scope_types=['system'],
        description='Accept sudo from user to tenant.',
        deprecated_rule=deprecated_use_sudo)
]


def list_rules():
    return rules



# Copyright 2015 Rackspace Inc.
#
# Author: Tim Simmons <tim.simmons@rackspace.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from designate import objects
from designate.objects.adapters.api_v2 import base


class ZoneExportAPIv2Adapter(base.APIv2Adapter):
    ADAPTER_OBJECT = objects.ZoneExport
    MODIFICATIONS = {
        'fields': {
            "id": {},
            "status": {},
            "message": {},
            "location": {},
            "zone_id": {},
            "project_id": {
                'rename': 'tenant_id'
            },
            "created_at": {},
            "updated_at": {},
            "version": {},
        },
        'options': {
            'links': True,
            'resource_name': 'export',
            'collection_name': 'exports',
        }
    }

    @classmethod
    def _get_path(cls, request, *args):
        return '/v2/zones/tasks/exports'

    @classmethod
    def render_object(cls, obj, *args, **kwargs):
        new_obj = super(ZoneExportAPIv2Adapter, cls).render_object(
            obj, *args, **kwargs
        )

        if (new_obj['location'] and
                new_obj['location'].startswith('designate://')):
            # Get the base uri from the self link, which respects host headers
            base_uri = new_obj['links']['self'].split(
                cls._get_path(kwargs['request']))[0]

            new_obj['links']['export'] = (
                    '%s/%s' % (base_uri, new_obj['location'].split('://')[1])
            )

        return new_obj


class ZoneExportListAPIv2Adapter(base.APIv2Adapter):
    ADAPTER_OBJECT = objects.ZoneExportList
    MODIFICATIONS = {
        'options': {
            'links': True,
            'resource_name': 'export',
            'collection_name': 'exports',
        }
    }



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from designate.notification_handler import base
from designate.tests import TestCase


class InheritFormBaseAddressHandler(base.BaseAddressHandler):
    """Class to inherit from BaseAddressHandler to test its methods

    Because BaseAddressHandler is an abstract class, in order to test methods
    we need to create something to inherit from it so we have something
    instantiatable.
    """
    def get_event_types(self):
        pass

    def get_exchange_topics(self):
        pass

    def process_notification(self):
        pass


class BaseAddressHandlerTest(TestCase):
    def test_get_ip_data_support_v6(self):
        addr_dict = {'address': '1762::B03:1:AF18', 'version': 6}
        baseaddresshandler = InheritFormBaseAddressHandler()
        observe = baseaddresshandler._get_ip_data(addr_dict)
        expect = {'octet1': 'B03', 'octet0': '1762', 'octet3': 'AF18',
                  'octet2': '1', 'ip_version': 6,
                  'ip_address': '1762--B03-1-AF18'}
        self.assertEqual(observe, expect)



# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import oslotest.base
import requests_mock

from designate.backend import impl_infoblox
from designate.backend.impl_infoblox import ibexceptions
from designate import context
from designate import exceptions
from designate import objects


class InfobloxBackendTestCase(oslotest.base.BaseTestCase):
    def setUp(self):
        super(InfobloxBackendTestCase, self).setUp()
        self.base_address = 'https://localhost/wapi'

        self.context = mock.Mock()
        self.admin_context = mock.Mock()
        mock.patch.object(
            context.DesignateContext, 'get_admin_context',
            return_value=self.admin_context).start()

        self.zone = objects.Zone(
            id='e2bed4dc-9d01-11e4-89d3-123b93f75cba',
            name='example.com.',
            email='example@example.com',
        )
        self.target = {
            'id': '4588652b-50e7-46b9-b688-a9bad40a873e',
            'type': 'infoblox',
            'masters': [
                {'host': '1.1.1.1', 'port': 53},
            ],
            'options': [
                {'key': 'wapi_url', 'value': 'https://localhost/wapi/v2.0/'},
                {'key': 'username', 'value': 'test'},
                {'key': 'password', 'value': 'test'},
                {'key': 'ns_group', 'value': 'test'},
            ]
        }

        self.backend = impl_infoblox.InfobloxBackend(
            objects.PoolTarget.from_dict(self.target)
        )

    @requests_mock.mock()
    def test_create_zone(self, req_mock):
        req_mock.post(
            '%s/v2.0/zone_auth' % self.base_address,
            json={},
        )

        req_mock.get(
            '%s/v2.0/zone_auth' % self.base_address,
            json={},
        )

        self.backend.create_zone(self.context, self.zone)

    def test_update_zone(self):
        self.backend.update_zone(self.context, self.zone)

    @requests_mock.mock()
    def test_delete_zone(self, req_mock):
        req_mock.post(
            '%s/v2.0/zone_auth' % self.base_address,
            json={},
        )

        req_mock.get(
            '%s/v2.0/zone_auth' % self.base_address,
            json={},
        )

        self.backend.create_zone(self.context, self.zone)
        self.backend.delete_zone(self.context, self.zone)

    def test_missing_wapi_url(self):
        target = dict(self.target)
        target['options'] = [
            {'key': 'username', 'value': 'test'},
            {'key': 'password', 'value': 'test'},
            {'key': 'ns_group', 'value': 'test'},
        ]

        pool_target = objects.PoolTarget.from_dict(target)

        self.assertRaisesRegex(
            ibexceptions.InfobloxIsMisconfigured, "wapi_url",
            impl_infoblox.InfobloxBackend, pool_target,
        )

    def test_missing_username(self):
        target = dict(self.target)
        target['options'] = [
            {'key': 'wapi_url', 'value': 'test'},
            {'key': 'password', 'value': 'test'},
            {'key': 'ns_group', 'value': 'test'}
        ]

        pool_target = objects.PoolTarget.from_dict(target)

        self.assertRaisesRegex(
            ibexceptions.InfobloxIsMisconfigured, "username",
            impl_infoblox.InfobloxBackend, pool_target,
        )

    def test_missing_password(self):
        target = dict(self.target)
        target['options'] = [
            {'key': 'wapi_url', 'value': 'test'},
            {'key': 'username', 'value': 'test'},
            {'key': 'ns_group', 'value': 'test'},
        ]

        pool_target = objects.PoolTarget.from_dict(target)

        self.assertRaisesRegex(
            ibexceptions.InfobloxIsMisconfigured, "password",
            impl_infoblox.InfobloxBackend, pool_target,
        )

    def test_missing_ns_group(self):
        target = dict(self.target)
        target['options'] = [
            {'key': 'wapi_url', 'value': 'test'},
            {'key': 'username', 'value': 'test'},
            {'key': 'password', 'value': 'test'},
        ]

        pool_target = objects.PoolTarget.from_dict(target)

        self.assertRaisesRegex(
            ibexceptions.InfobloxIsMisconfigured, "ns_group",
            impl_infoblox.InfobloxBackend, pool_target,
        )

    def test_wrong_port(self):
        target = dict(self.target)
        target['masters'] = [
            {'host': '1.1.1.1', 'port': 100},
        ]

        pool_target = objects.PoolTarget.from_dict(target)

        self.assertRaisesRegex(
            exceptions.ConfigurationError,
            'Infoblox only supports mDNS instances on port 53',
            impl_infoblox.InfobloxBackend, pool_target,
        )



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Endre Karlson <endre.karlson@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

import dns
import dns.query
from oslo_config import cfg

from designate import dnsutils
from designate import exceptions
from designate import objects
import designate.tests

CONF = cfg.CONF
SAMPLES = {
    ("cname.example.com.", "CNAME"): {
        "ttl": 10800,
        "records": ["example.com."],
    },
    ("_http._tcp.example.com.", "SRV"): {
        "ttl": 10800,
        "records": [
            "10 0 80 192.0.0.4.example.com.",
            "10 5 80 192.0.0.5.example.com."
        ],
    },
    ("ipv4.example.com.", "A"): {
        "ttl": 300,
        "records": ["192.0.0.1"]
    },
    ("delegation.example.com.", "NS"): {
        "ttl": 10800,
        "records": ["ns1.example.com."]
    },
    ("ipv6.example.com.", "AAAA"): {
        "ttl": 10800,
        "records": ["fd00::1"],
    },
    ("example.com.", "SOA"): {
        "records": [
            "ns1.example.com. nsadmin.example.com."
            " 2013091101 7200 3600 2419200 10800"
        ],
        "ttl": 600
    },
    ("example.com.", "MX"): {
        "ttl": 10800,
        "records": [
            "5 192.0.0.2.example.com.",
            '10 192.0.0.3.example.com.'
        ]
    },
    ("example.com.", "TXT"): {
        "ttl": 10800,
        "records": ['"abc" "def"']
    },
    ("example.com.", "SPF"): {
        "ttl": 10800,
        "records": ['"v=spf1 mx a"']
    },
    ("example.com.", "NS"): {
        "ttl": 10800,
        "records": [
            'ns1.example.com.',
            'ns2.example.com.'
        ]
    }
}


class TestUtils(designate.tests.TestCase):
    def setUp(self):
        super(TestUtils, self).setUp()

    def test_from_dnspython_zone(self):
        zone_file = self.get_zonefile_fixture()

        dnspython_zone = dns.zone.from_text(
            zone_file,
            relativize=False,
            check_origin=False
        )

        zone = dnsutils.from_dnspython_zone(dnspython_zone)

        self.assertIsInstance(zone, objects.zone.Zone)

    def test_from_dnspython_zone_no_soa(self):
        zone_file = self.get_zonefile_fixture(variant='nosoa')

        dnspython_zone = dns.zone.from_text(
            zone_file,
            relativize=False,
            check_origin=False
        )

        self.assertRaisesRegex(
            exceptions.BadRequest,
            'An SOA record is required',
            dnsutils.from_dnspython_zone, dnspython_zone,
        )

    def test_parse_zone(self):
        zone_file = self.get_zonefile_fixture()

        dnspython_zone = dns.zone.from_text(
            zone_file,
            # Don't relativize, otherwise we end up with '@' record names.
            relativize=False,
            # Dont check origin, we allow missing NS records (missing SOA
            # records are taken care of in _create_zone).
            check_origin=False
        )

        zone = dnsutils.from_dnspython_zone(dnspython_zone)

        for rrset in zone.recordsets:
            k = (rrset.name, rrset.type)
            self.assertIn(k, SAMPLES)

            sample_ttl = SAMPLES[k].get('ttl', None)
            if rrset.obj_attr_is_set('ttl') or sample_ttl is not None:
                self.assertEqual(sample_ttl, rrset.ttl)

            self.assertEqual(len(rrset.records), len(SAMPLES[k]['records']))

            for record in rrset.records:
                self.assertIn(record.data, SAMPLES[k]['records'])

        self.assertEqual(len(SAMPLES), len(zone.recordsets))
        self.assertEqual('example.com.', zone.name)

    def test_zone_lock(self):
        # Initialize a ZoneLock
        lock = dnsutils.ZoneLock(0.1)

        # Ensure there's no lock for different zones
        for zone_name in ['foo.com.', 'bar.com.', 'example.com.']:
            self.assertTrue(lock.acquire(zone_name))

        # Ensure a lock for successive calls for the same zone
        self.assertTrue(lock.acquire('example2.com.'))
        self.assertFalse(lock.acquire('example2.com.'))

        # Acquire, release, and reacquire
        self.assertTrue(lock.acquire('example3.com.'))
        lock.release('example3.com.')
        self.assertTrue(lock.acquire('example3.com.'))

    def test_limit_notify_middleware(self):
        self.CONF.set_override('notify_delay', 0.1, 'service:agent')

        # Initialize the middlware
        placeholder_app = None
        middleware = dnsutils.LimitNotifyMiddleware(placeholder_app)

        # Prepare a NOTIFY
        zone_name = 'example.com.'
        notify = dns.message.make_query(zone_name, dns.rdatatype.SOA)
        notify.flags = 0
        notify.set_opcode(dns.opcode.NOTIFY)
        notify.flags |= dns.flags.AA

        # Send the NOTIFY through the middleware
        # No problem, middleware should return None to pass it on
        self.assertIsNone(middleware.process_request(notify))

    @mock.patch('designate.dnsutils.ZoneLock.acquire', return_value=False)
    def test_limit_notify_middleware_no_acquire(self, mock_acquire):
        self.CONF.set_override('notify_delay', 0.1, 'service:agent')

        # Initialize the middlware
        placeholder_app = None
        middleware = dnsutils.LimitNotifyMiddleware(placeholder_app)

        # Prepare a NOTIFY
        zone_name = 'example.com.'
        notify = dns.message.make_query(zone_name, dns.rdatatype.SOA)
        notify.flags = 0
        notify.set_opcode(dns.opcode.NOTIFY)
        notify.flags |= dns.flags.AA

        # Make a response object to match the middleware's return
        response = dns.message.make_response(notify)
        # Provide an authoritative answer
        response.flags |= dns.flags.AA

        # Send the NOTIFY through the middleware
        # Lock can't be acquired, a NOTIFY is already being worked on
        # so just return what would have come back for a successful NOTIFY
        # This needs to be a one item tuple for the serialization middleware
        self.assertEqual(middleware.process_request(notify), (response,))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import oslotest.base
import stevedore.exception

from designate import backend
from designate.backend import base
from designate.backend import impl_pdns4
from designate import context
from designate import objects
from designate.tests import fixtures


class BaseBackendTestCase(oslotest.base.BaseTestCase):
    def setUp(self):
        super(BaseBackendTestCase, self).setUp()
        self.stdlog = fixtures.StandardLogging()
        self.useFixture(self.stdlog)

        self.context = mock.Mock()
        self.admin_context = mock.Mock()
        mock.patch.object(
            context.DesignateContext, 'get_admin_context',
            return_value=self.admin_context).start()

        self.target = {
            'type': 'pdns4',
            'masters': [
            ],
            'options': [
            ],
        }

    @mock.patch.object(base.Backend, 'get_driver')
    def test_untested_backend(self, mock_get_driver):
        driver = mock.Mock()
        driver.__backend_status__ = 'untested'
        mock_get_driver.return_value = driver

        self.target['type'] = 'test'
        pool_target = objects.PoolTarget.from_dict(self.target)

        backend.get_backend(pool_target)

        self.assertIn('WARNING', self.stdlog.logger.output)
        self.assertIn(
            "Backend Driver 'test' loaded. Has status of 'untested'",
            self.stdlog.logger.output
        )

    @mock.patch.object(base.Backend, 'get_driver')
    def test_tested_backend(self, mock_get_driver):
        driver = mock.Mock()
        driver.__backend_status__ = 'integrated'
        mock_get_driver.return_value = driver

        self.target['type'] = 'test'
        pool_target = objects.PoolTarget.from_dict(self.target)

        backend.get_backend(pool_target)

        self.assertNotIn('WARNING', self.stdlog.logger.output)
        self.assertIn(
            "Backend Driver 'test' loaded. Has status of 'integrated'",
            self.stdlog.logger.output
        )

    def test_get_backend(self):
        pool_target = objects.PoolTarget.from_dict(self.target)
        self.assertIsInstance(
            backend.get_backend(pool_target),
            impl_pdns4.PDNS4Backend
        )

    def test_get_backend_does_not_exist(self):
        self.target['type'] = 'unknown'
        pool_target = objects.PoolTarget.from_dict(self.target)
        self.assertRaises(
            stevedore.exception.NoMatches,
            backend.get_backend, pool_target
        )



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging

from designate.schema import format
from designate.tests import TestCase


LOG = logging.getLogger(__name__)


class SchemaFormatTest(TestCase):
    def test_is_ipv4(self):
        valid_ipaddresses = [
            '0.0.0.1',
            '127.0.0.1',
            '10.0.0.1',
            '192.0.2.2',
        ]

        invalid_ipaddresses = [
            '0.0.0.0',
            '0.0.0.256',
            '0.0.256.0',
            '0.256.0.0',
            '256.0.0.0',
            '127.0.0',
            '127.0.0.999',
            '127.0.0.256',
            '127.0..1',
            '-1.0.0.1',
            '1.0.-0.1',
            '1.0.0.-1',
            'ABCDEF',
            'ABC/DEF',
            'ABC\\DEF',
            # Trailing newline - Bug 1471158
            "127.0.0.1\n",
        ]

        for ipaddress in valid_ipaddresses:
            self.assertTrue(format.is_ipv4(ipaddress))

        for ipaddress in invalid_ipaddresses:
            self.assertFalse(format.is_ipv4(ipaddress))

    def test_is_ipv6(self):
        valid_ipaddresses = [
            '2001:db8::0',
            '2001:0db8:85a3:0000:0000:8a2e:0370:7334',
            '2001:db8:85a3:0000:0000:8a2e:0370:7334',
            '2001:db8:85a3::8a2e:0370:7334',
        ]

        invalid_ipaddresses = [
            # Invalid characters
            'hhhh:hhhh:hhhh:hhhh:hhhh:hhhh:hhhh:hhhh'
            # Trailing newline - Bug 1471158
            "2001:db8::0\n",
        ]

        for ipaddress in valid_ipaddresses:
            self.assertTrue(format.is_ipv6(ipaddress),
                            'Expected Valid: %s' % ipaddress)

        for ipaddress in invalid_ipaddresses:
            self.assertFalse(format.is_ipv6(ipaddress),
                             'Expected Invalid: %s' % ipaddress)

    def test_is_hostname(self):
        valid_hostnames = [
            'example.com.',
            'www.example.com.',
            '*.example.com.',
            '12345.example.com.',
            '192-0-2-1.example.com.',
            'ip192-0-2-1.example.com.',
            'www.ip192-0-2-1.example.com.',
            'ip192-0-2-1.www.example.com.',
            'abc-123.example.com.',
            '_tcp.example.com.',
            '_service._tcp.example.com.',
            ('1.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.8.b.d.0.1.0.0.2'
             '.ip6.arpa.'),
            '1.1.1.1.in-addr.arpa.',
            'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.',
            ('abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghi.'),
        ]

        invalid_hostnames = [
            '**.example.com.',
            '*.*.example.org.',
            'a.*.example.org.',
            # Exceeds single lable length limit
            ('abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijkL'
             '.'),
            ('abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijkL'
             '.'),
            # Exceeds total length limit
            ('abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopq.'),
            # Empty label part
            'abc..def.',
            '..',
            # Invalid character
            'abc$.def.',
            'abc.def$.',
            # Labels must not start with a -
            '-abc.',
            'abc.-def.',
            'abc.-def.ghi.',
            # Labels must not end with a -
            'abc-.',
            'abc.def-.',
            'abc.def-.ghi.',
            # Labels must not start or end with a -
            '-abc-.',
            'abc.-def-.',
            'abc.-def-.ghi.',
            # Trailing newline - Bug 1471158
            "www.example.com.\n",
        ]

        for hostname in valid_hostnames:
            self.assertTrue(format.is_hostname(hostname))

        for hostname in invalid_hostnames:
            self.assertFalse(format.is_hostname(hostname))

    def test_is_ns_hostname(self):
        valid_ns_hostnames = [
            'example.com.',
            'www.example.com.',
            '12345.example.com.',
            '192-0-2-1.example.com.',
            'ip192-0-2-1.example.com.',
            'www.ip192-0-2-1.example.com.',
            'ip192-0-2-1.www.example.com.',
            'abc-123.example.com.',
            '_tcp.example.com.',
            '_service._tcp.example.com.',
            ('1.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.8.b.d.0.1.0.0.2'
             '.ip6.arpa.'),
            '1.1.1.1.in-addr.arpa.',
            'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.',
            ('abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghi.'),
        ]

        invalid_ns_hostnames = [
            # Wildcard NS hostname, bug #1533299
            '*.example.com.',
            '**.example.com.',
            '*.*.example.org.',
            'a.*.example.org.',
            # Exceeds single lable length limit
            ('abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijkL'
             '.'),
            ('abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijkL'
             '.'),
            # Exceeds total length limit
            ('abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopq.'),
            # Empty label part
            'abc..def.',
            '..',
            # Invalid character
            'abc$.def.',
            'abc.def$.',
            # Labels must not start with a -
            '-abc.',
            'abc.-def.',
            'abc.-def.ghi.',
            # Labels must not end with a -
            'abc-.',
            'abc.def-.',
            'abc.def-.ghi.',
            # Labels must not start or end with a -
            '-abc-.',
            'abc.-def-.',
            'abc.-def-.ghi.',
            # Trailing newline - Bug 1471158
            "www.example.com.\n",
        ]

        for hostname in valid_ns_hostnames:
            self.assertTrue(format.is_ns_hostname(hostname))

        for hostname in invalid_ns_hostnames:
            self.assertFalse(format.is_ns_hostname(hostname))

    def test_is_zonename(self):
        valid_zonenames = [
            'example.com.',
            'www.example.com.',
            '12345.example.com.',
            '192-0-2-1.example.com.',
            'ip192-0-2-1.example.com.',
            'www.ip192-0-2-1.example.com.',
            'ip192-0-2-1.www.example.com.',
            'abc-123.example.com.',
            '_tcp.example.com.',
            '_service._tcp.example.com.',
            ('1.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.8.b.d.0.1.0.0.2'
             '.ip6.arpa.'),
            '1.1.1.1.in-addr.arpa.',
            'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.',
            ('abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghi.'),
        ]

        invalid_zonenames = [
            '*.example.com.',
            '**.example.com.',
            '*.*.example.org.',
            'a.*.example.org.',
            # Exceeds single lable length limit
            ('abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijkL'
             '.'),
            ('abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijkL'
             '.'),
            ('abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijkL'
             '.'),
            # Exceeds total length limit
            ('abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopq.'),
            # Empty label part
            'abc..def.',
            '..',
            # Invalid character
            'abc$.def.',
            'abc.def$.',
            # Labels must not start with a -
            '-abc.',
            'abc.-def.',
            'abc.-def.ghi.',
            # Labels must not end with a -
            'abc-.',
            'abc.def-.',
            'abc.def-.ghi.',
            # Labels must not start or end with a -
            '-abc-.',
            'abc.-def-.',
            'abc.-def-.ghi.',
            # Trailing newline - Bug 1471158
            "example.com.\n",
        ]

        for zonename in valid_zonenames:
            self.assertTrue(format.is_zonename(zonename), zonename)

        for zonename in invalid_zonenames:
            self.assertFalse(format.is_zonename(zonename), zonename)

    def test_is_srv_hostname(self):
        valid_hostnames = [
            '_sip._tcp.example.com.',
            '_sip._udp.example.com.',
        ]

        invalid_hostnames = [
            # Invalid Formats
            '_tcp.example.com.',
            'sip._udp.example.com.',
            '_sip.udp.example.com.',
            'sip.udp.example.com.',
            # Trailing newline - Bug 1471158
            "_sip._tcp.example.com.\n",
        ]

        for hostname in valid_hostnames:
            self.assertTrue(format.is_srv_hostname(hostname),
                            'Expected Valid: %s' % hostname)

        for hostname in invalid_hostnames:
            self.assertFalse(format.is_srv_hostname(hostname),
                             'Expected Invalid: %s' % hostname)

    def test_is_tldname(self):
        valid_tldnames = [
            'com',
            'net',
            'org',
            'co.uk',
        ]

        invalid_tldnames = [
            # Invalid Formats
            'com.',
            '.com',
            # Trailing newline - Bug 1471158
            "com\n",
        ]

        for tldname in valid_tldnames:
            self.assertTrue(format.is_tldname(tldname),
                            'Expected Valid: %s' % tldname)

        for tldname in invalid_tldnames:
            self.assertFalse(format.is_tldname(tldname),
                             'Expected Invalid: %s' % tldname)

    def test_is_email(self):
        valid_emails = [
            'user@example.com',
            'user@emea.example.com',
            'user@example.com',
            'first.last@example.com',
        ]

        invalid_emails = [
            # We use the email addr for the SOA RNAME field, this means the
            # entire address, excluding the @ must be chacracters valid
            # as a DNS name. i.e. + and % addressing is invalid.
            'user+plus@example.com',
            'user%example.org@example.com',
            'example.org',
            '@example.org',
            'user@*.example.org',
            'user',
            'user@',
            'user+plus',
            'user+plus@',
            'user%example.org',
            'user%example.org@',
            'user@example.org.',
            # Exceeds total length limit
            ('user@fghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
             'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopq.'),
            # Exceeds single lable length limit
            ('user@abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefg'
             'hijkL.'),
            ('user@abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefg'
             'hijk.abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefg'
             'hijkL.'),
            # Exceeds single lable length limit in username part
            ('abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijkL'
             '@example.com.'),
        ]

        for email in valid_emails:
            LOG.debug('Expecting success for: %s' % email)
            self.assertTrue(format.is_email(email))

        for email in invalid_emails:
            LOG.debug('Expecting failure for: %s' % email)
            self.assertFalse(format.is_email(email))

    def test_is_sshfp(self):
        valid_sshfps = [
            '72d30d211ce8c464de2811e534de23b9be9b4dc4',
            '7f3f61e323a7d75de08a2a6069b333e925cae260f4902017194002f226db8658',
        ]

        invalid_sshfps = [
            # Invalid Formats
            'P2d30d211ce8c464de2811e534de23b9be9b4dc4',  # "P" !IN [A-F]
            '72d30d211',
            # Trailing newline - Bug 1471158
            "72d30d211ce8c464de2811e534de23b9be9b4dc4\n",
        ]

        for sshfp in valid_sshfps:
            self.assertTrue(format.is_sshfp_fingerprint(sshfp),
                            'Expected Valid: %s' % sshfp)

        for sshfp in invalid_sshfps:
            self.assertFalse(format.is_sshfp_fingerprint(sshfp),
                             'Expected Invalid: %s' % sshfp)

    def test_is_uuid(self):
        valid_uuids = [
            'd3693ef8-2188-11e5-bf77-676ff9eb39dd',
        ]

        invalid_uuids = [
            # Invalid Formats
            'p3693ef8-2188-11e5-bf77-676ff9eb39dd',  # "p" !IN [A-F]
            'd3693ef8218811e5bf77676ff9eb39dd',
            # Trailing newline - Bug 1471158
            "d3693ef8-2188-11e5-bf77-676ff9eb39dd\n",
        ]

        for uuid in valid_uuids:
            self.assertTrue(format.is_uuid(uuid),
                            'Expected Valid: %s' % uuid)

        for uuid in invalid_uuids:
            self.assertFalse(format.is_uuid(uuid),
                             'Expected Invalid: %s' % uuid)

    def test_is_fip_id(self):
        valid_fip_ids = [
            'region-a:d3693ef8-2188-11e5-bf77-676ff9eb39dd',
        ]

        invalid_fip_ids = [
            # Invalid Formats
            'region-a:p3693ef8-2188-11e5-bf77-676ff9eb39dd',  # "p" !IN [A-F]
            # Trailing newline - Bug 1471158
            "region-a:d3693ef8-2188-11e5-bf77-676ff9eb39dd\n",
        ]

        for fip_id in valid_fip_ids:
            self.assertTrue(format.is_floating_ip_id(fip_id),
                            'Expected Valid: %s' % fip_id)

        for fip_id in invalid_fip_ids:
            self.assertFalse(format.is_floating_ip_id(fip_id),
                             'Expected Invalid: %s' % fip_id)

    def test_is_ip_and_port(self):
        valid_ip_and_ports = [
            '192.0.2.1:80',
            '192.0.2.1:1',
            '192.0.2.1:65535',
        ]

        invalid_ip_and_ports = [
            '192.0.2.1:65536',
            # Trailing newline - Bug 1471158
            "192.0.2.1:80\n",
        ]

        for ip_and_port in valid_ip_and_ports:
            self.assertTrue(format.is_ip_and_port(ip_and_port),
                            'Expected Valid: %s' % ip_and_port)

        for ip_and_port in invalid_ip_and_ports:
            self.assertFalse(format.is_ip_and_port(ip_and_port),
                             'Expected Invalid: %s' % ip_and_port)



# Copyright 2016 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging
from oslo_utils.strutils import bool_from_string

from designate import exceptions
from designate.objects import PoolList
from designate.scheduler.filters import base

LOG = logging.getLogger(__name__)


class AttributeFilter(base.Filter):
    """This allows users to choose the pool by supplying hints to this filter.
    These are provided as attributes as part of the zone object provided at
    zone create time.

    .. code-block:: javascript
        :emphasize-lines: 3,4,5

        {
            "attributes": {
                "pool_level": "gold",
                "fast_ttl": "true",
                "pops": "global",
            },
            "email": "user@example.com",
            "name": "example.com."
        }

    The zone attributes are matched against the potential pool candidates, and
    any pools that do not match **all** hints are removed.

    .. warning::

        This should be uses in conjunction with the
        :class:`designate.scheduler.impl_filter.filters.random_filter.RandomFilter`
        in case of multiple Pools matching the filters, as without it, we will
        raise an error to the user.
        """

    name = 'attribute'
    """Name to enable in the ``[designate:central:scheduler].filters`` option
    list
    """

    def filter(self, context, pools, zone):

        try:
            zone_attributes = zone.attributes.to_dict()
        except exceptions.RelationNotLoaded:
            zone_attributes = {}

        def evaluate_pool(pool):
            try:
                pool_attributes = pool.attributes.to_dict()
            except exceptions.RelationNotLoaded:
                pool_attributes = {}

            # Remove the "pool_id" attribute, that is used in
            # PoolIDAttributeFilter. If the item is not in the dict, it is
            # fine, we should just continue.
            pool_attributes.pop('pool_id', None)

            if not pool_attributes:
                # If we did not send any attribute to filter on, we should
                # not filter the pools based on an empty set, as this will
                # return no pools.
                return True

            # Check if the keys requested exist in this pool
            if not {key for key in pool_attributes.keys()}.issuperset(
                    zone_attributes):
                LOG.debug(
                    '%(pool)s did not match list of requested attribute '
                    'keys - removing from list. Requested: %(r_key)s. Pool:'
                    '%(p_key)s',
                    {
                        'pool': pool,
                        'r_key': zone_attributes,
                        'p_key': pool_attributes
                    }
                )
                # Missing required keys - remove from the list
                return False

            for key in zone_attributes.keys():
                LOG.debug('Checking value of %s for %s', key, pool)

                pool_attr = bool_from_string(pool_attributes[key],
                                             default=pool_attributes[key])
                zone_attr = bool_from_string(zone_attributes[key],
                                             default=zone_attributes[key])

                if not pool_attr == zone_attr:
                    LOG.debug(
                        '%(pool)s did not match requested value of %(key)s. '
                        'Requested: %(r_val)s. Pool: %(p_val)s',
                        {
                            'pool': pool,
                            'key': key,
                            'r_val': zone_attr,
                            'p_val': pool_attr
                        })
                    # Value didn't match - remove from the list
                    return False

            # Pool matches list of attributes - keep
            return True

        pool_list = [pool for pool in pools if evaluate_pool(pool)]
        pools = PoolList(objects=pool_list)
        return pools



# Copyright 2015 Rackspace Inc.
#
# Author: Tim Simmons <tim.simmons@rackspae.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging
import pecan

from designate.api.v2.controllers import rest
from designate import exceptions
from designate.objects.adapters import DesignateAdapter
from designate import utils

LOG = logging.getLogger(__name__)


class ZoneImportController(rest.RestController):

    SORT_KEYS = ['created_at', 'id', 'updated_at']

    @pecan.expose(template='json:', content_type='application/json')
    @utils.validate_uuid('import_id')
    def get_one(self, import_id):
        """Get Zone Imports"""

        request = pecan.request
        context = request.environ['context']

        zone_import = self.central_api.get_zone_import(
            context, import_id)

        LOG.info("Retrieved %(import)s", {'import': zone_import})

        return DesignateAdapter.render('API_v2', zone_import, request=request)

    @pecan.expose(template='json:', content_type='application/json')
    def get_all(self, **params):
        """List Zone Imports"""
        request = pecan.request
        context = request.environ['context']
        marker, limit, sort_key, sort_dir = utils.get_paging_params(
                context, params, self.SORT_KEYS)

        # Extract any filter params.
        accepted_filters = ('status', 'message', 'zone_id', )

        criterion = self._apply_filter_params(
            params, accepted_filters, {})

        zone_imports = self.central_api.find_zone_imports(
            context, criterion, marker, limit, sort_key, sort_dir)

        LOG.info("Retrieved %(imports)s", {'imports': zone_imports})

        return DesignateAdapter.render('API_v2', zone_imports, request=request)

    @pecan.expose(template='json:', content_type='application/json')
    def post_all(self):
        """Create Zone Import"""
        request = pecan.request
        response = pecan.response
        context = request.environ['context']
        body = request.body.decode('utf-8')

        if request.content_type != 'text/dns':
            raise exceptions.UnsupportedContentType(
                'Content-type must be text/dns')

        # Create the zone_import
        zone_import = self.central_api.create_zone_import(
            context, body)
        response.status_int = 202

        LOG.info("Created %(zone_import)s", {'zone_import': zone_import})

        zone_import = DesignateAdapter.render('API_v2', zone_import,
                                              request=request)

        response.headers['Location'] = zone_import['links']['self']
        # Prepare and return the response body
        return zone_import

    # NOTE: template=None is important here, template='json:' manifests
    #       in this bug: https://bugs.launchpad.net/designate/+bug/1592153
    @pecan.expose(template=None, content_type='application/json')
    @utils.validate_uuid('zone_import_id')
    def delete_one(self, zone_import_id):
        """Delete Zone Import"""
        request = pecan.request
        response = pecan.response
        context = request.environ['context']

        zone_import = self.central_api.delete_zone_import(
            context, zone_import_id)

        LOG.info("Deleted %(zone_import)s", {'zone_import': zone_import})

        response.status_int = 204

        return ''



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_config import cfg

API_GROUP = cfg.OptGroup(
    name='service:api',
    title="Configuration for API Service"
)

API_OPTS = [
    cfg.IntOpt('workers',
               help='Number of api worker processes to spawn'),
    cfg.IntOpt('threads', default=1000,
               help='Number of api greenthreads to spawn'),
    cfg.BoolOpt('enable_host_header', default=True,
                help='Enable host request headers'),
    cfg.StrOpt('api_base_uri', default='http://127.0.0.1:9001/',
               help='the url used as the base for all API responses,'
                    'This should consist of the scheme (http/https),'
                    'the hostname, port, and any paths that are added'
                    'to the base of Designate is URLs,'
                    'For example http://dns.openstack.example.com/dns'),
    cfg.ListOpt('listen',
                default=['0.0.0.0:9001'],
                help='API host:port pairs to listen on'),
    cfg.StrOpt('api_paste_config', default='api-paste.ini',
               help='File name for the paste.deploy config for designate-api'),
    cfg.StrOpt('auth_strategy', default='keystone',
               help='The strategy to use for auth. Supports noauth or '
                    'keystone'),
    cfg.BoolOpt('enable_api_v2', default=True,
                help='Enable the Designate V2 API'),
    cfg.BoolOpt('enable_api_admin', default=False,
                help='enable-api-admin'),
    cfg.IntOpt('max_header_line', default=16384,
               deprecated_for_removal=True,
               deprecated_reason='This parameter has had no effect since '
                                 'Train. Use [DEFAULT] max_header_line '
                                 'instead.',
               help="Maximum line size of message headers to be accepted. "
                    "max_header_line may need to be increased when using "
                    "large tokens (typically those generated by the "
                    "Keystone v3 API with big service catalogs)."),
    cfg.BoolOpt('pecan_debug', default=False,
                help='Pecan HTML Debug Interface'),
]

APT_V2_OPTS = [
    cfg.ListOpt('enabled_extensions_v2', default=[],
                help='Enabled API Extensions for the V2 API'),
    cfg.IntOpt('default_limit_v2', default=20,
               help='Default per-page limit for the V2 API, a value of None '
                    'means show all results by default'),
    cfg.IntOpt('max_limit_v2', default=1000,
               help='Max per-page limit for the V2 API'),
    cfg.BoolOpt('quotas_verify_project_id', default=False,
                help='Verify that the requested Project ID for quota target '
                     'is a valid project in Keystone.'),
]

API_ADMIN_OPTS = [
    cfg.ListOpt('enabled_extensions_admin', default=[],
                help='Enabled Admin API Extensions'),
    cfg.IntOpt('default_limit_admin', default=20,
               help='Default per-page limit for the Admin API, a value of None'
                    ' means show all results by default'),
    cfg.IntOpt('max_limit_admin', default=1000,
               help='Max per-page limit for the Admin API'),
]

API_MIDDLEWARE_OPTS = [
    cfg.BoolOpt('maintenance_mode', default=False,
                help='Enable API Maintenance Mode'),
    cfg.StrOpt('maintenance_mode_role', default='admin',
               help='Role allowed to bypass maintaince mode'),
    cfg.StrOpt('secure_proxy_ssl_header',
               default='X-Forwarded-Proto',
               deprecated_for_removal=True,
               deprecated_reason='This parameter has had no effect since '
                                 'SSLMiddleware was replaced by '
                                 'http_proxy_to_wsgi',
               help="The HTTP Header that will be used to determine which "
                    "the original request protocol scheme was, even if it was "
                    "removed by an SSL terminating proxy."),
    cfg.StrOpt('override_proto',
               deprecated_for_removal=True,
               deprecated_reason='This parameter has had no effect since '
                                 'SSLMiddleware was replaced by '
                                 'http_proxy_to_wsgi',
               help="A scheme that will be used to override "
                    "the request protocol scheme, even if it was "
                    "set by an SSL terminating proxy."),
]


def register_opts(conf):
    conf.register_group(API_GROUP)
    conf.register_opts(API_OPTS, group=API_GROUP)
    conf.register_opts(APT_V2_OPTS, group=API_GROUP)
    conf.register_opts(API_ADMIN_OPTS, group=API_GROUP)
    conf.register_opts(API_MIDDLEWARE_OPTS, group=API_GROUP)


def list_opts():
    return {
        API_GROUP: (API_OPTS +
                    APT_V2_OPTS +
                    API_ADMIN_OPTS +
                    API_MIDDLEWARE_OPTS)
    }



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from designate import objects
from designate.objects.adapters.api_v2 import base


class FloatingIPAPIv2Adapter(base.APIv2Adapter):
    ADAPTER_OBJECT = objects.FloatingIP
    MODIFICATIONS = {
        'fields': {
            "id": {
                'rename': 'key'
            },
            "description": {
                'read_only': False
            },
            "address": {},
            "ptrdname": {
                'read_only': False
            },
            "ttl": {
                'read_only': False
            },
            "action": {
                "read_only": True,
            },
            "status": {
                "read_only": True
            }
        },
        'options': {
            'links': True,
            'resource_name': 'floatingip',
            'collection_name': 'floatingips',
        }
    }

    @classmethod
    def _get_resource_links(cls, obj, request):
        return {
            'self': '%s%s/%s' % (
                cls._get_base_url(request),
                cls._get_path(request),
                obj.key
            )
        }


class FloatingIPListAPIv2Adapter(base.APIv2Adapter):
    ADAPTER_OBJECT = objects.FloatingIPList
    MODIFICATIONS = {
        'options': {
            'links': True,
            'resource_name': 'floatingip',
            'collection_name': 'floatingips',
        }
    }



# Copyright 2018 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

from oslo_upgradecheck import upgradecheck
from sqlalchemy.schema import MetaData
from sqlalchemy.schema import Table

from designate.cmd import status
from designate.sqlalchemy import session
from designate import tests


class TestDuplicateServiceStatus(tests.TestCase):
    def setUp(self):
        super(TestDuplicateServiceStatus, self).setUp()
        self.engine = session.get_engine('storage:sqlalchemy')
        self.meta = MetaData()
        self.meta.bind = self.engine
        self.service_statuses_table = Table('service_statuses', self.meta,
                                            autoload=True)

    def test_success(self):
        fake_record = {'id': '1',
                       'service_name': 'worker',
                       'hostname': 'localhost',
                       'status': 'UP',
                       'stats': '',
                       'capabilities': '',
                       }
        self.service_statuses_table.insert().execute(fake_record)
        # Different hostname should be fine
        fake_record['id'] = '2'
        fake_record['hostname'] = 'otherhost'
        self.service_statuses_table.insert().execute(fake_record)
        # Different service_name should be fine
        fake_record['id'] = '3'
        fake_record['service_name'] = 'producer'
        self.service_statuses_table.insert().execute(fake_record)
        checks = status.Checks()
        self.assertEqual(upgradecheck.Code.SUCCESS,
                         checks._duplicate_service_status().code)

    @mock.patch('designate.sqlalchemy.session.get_engine')
    def test_failure(self, mock_get_engine):
        mock_engine = mock.MagicMock()
        mock_execute = mock.MagicMock()
        mock_engine.execute.return_value = mock_execute
        mock_execute.fetchall.return_value = [(2,)]
        mock_get_engine.return_value = mock_engine

        checks = status.Checks()
        self.assertEqual(upgradecheck.Code.FAILURE,
                         checks._duplicate_service_status().code)



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import re

from designate import objects
from designate.objects.adapters.api_v2 import base


class NotSpecifiedSential:
    pass


REQUIRED_RE = re.compile(r"\'([\w]*)\' is a required property")


class ValidationErrorAPIv2Adapter(base.APIv2Adapter):
    ADAPTER_OBJECT = objects.ValidationError
    MODIFICATIONS = {
        'fields': {
            "path": {},
            "message": {},
            "validator": {},
            "validator_value": {},
        },
        'options': {
            'links': False,
            'resource_name': 'error',
            'collection_name': 'errors',
        }
    }

    @classmethod
    def render_object(cls, error, *args, **kwargs):
        # Do the usual rename
        error_dict = super(ValidationErrorAPIv2Adapter, cls).render_object(
            error, *args, **kwargs)

        # Currently JSON Schema doesn't add the path on for required items
        if error_dict.get('validator', '') == 'required':
            error_dict['path'].append(
                REQUIRED_RE.match(error.message).group(1))

        obj = kwargs['failed_object']
        # Rename the keys in the path list
        error_dict['path'] = cls._rename_path(obj, error_dict['path'])

        return error_dict

    @classmethod
    def _rename_path(cls, object, path):
        new_path = list()

        obj_adapter = cls.get_object_adapter(object)

        for path_segment in path:
            new_path_segment, obj_adapter = cls._rename_path_segment(
                obj_adapter, object, path_segment
            )
            new_path.append(new_path_segment)

        return new_path

    @classmethod
    def _rename_path_segment(cls, obj_adapter, obj, path_segment):

        # Check if the object is a list - lists will just have an index as a
        # value, ands this can't be renamed
        if issubclass(obj_adapter.ADAPTER_OBJECT, objects.ListObjectMixin):
            obj_adapter = cls.get_object_adapter(
                obj_adapter.ADAPTER_OBJECT.LIST_ITEM_TYPE.obj_name()
            )
            # Return the segment as is, and the next adapter (which is the
            # LIST_ITEM_TYPE)
            return path_segment, obj_adapter

        for key, value in obj_adapter.MODIFICATIONS.get(
                'fields', {}).items():

            # Check if this field as actually a nested object
            field = obj.FIELDS.get(path_segment, {})
            if isinstance(field, dict) and field.get('relation'):
                obj_cls = obj.FIELDS.get(path_segment).get('relation_cls')
                obj_adapter = cls.get_object_adapter(obj_cls)

                new_obj = objects.DesignateObject.obj_cls_from_name(obj_cls)()
                # Recurse down into this object
                path_segment, obj_adapter = cls._rename_path_segment(
                    obj_adapter, new_obj, path_segment
                )

                # No need to continue the loop
                break
            elif hasattr(field, 'objname'):
                obj_cls = field.objname
                obj_adapter = cls.get_object_adapter(obj_cls)

                new_obj = objects.DesignateObject.obj_cls_from_name(obj_cls)()
                # Recurse down into this object
                path_segment, obj_adapter = cls._rename_path_segment(
                    obj_adapter, new_obj, path_segment)

                # No need to continue the loop
                break

            if not isinstance(
                value.get(
                    'rename', NotSpecifiedSential()), NotSpecifiedSential)\
                    and path_segment == value.get('rename'):
                # Just do the rename
                path_segment = key

        return path_segment, obj_adapter


class ValidationErrorListAPIv2Adapter(base.APIv2Adapter):
    ADAPTER_OBJECT = objects.ValidationErrorList
    MODIFICATIONS = {
        'options': {
            'links': False,
            'resource_name': 'error',
            'collection_name': 'errors',
        }
    }



# Copyright 2012 Hewlett-Packard Development Company, L.P. All Rights Reserved.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_config import cfg

CENTRAL_GROUP = cfg.OptGroup(
    name='service:central',
    title="Configuration for Central Service"
)

CENTRAL_OPTS = [
    cfg.IntOpt('workers',
               help='Number of central worker processes to spawn'),
    cfg.IntOpt('threads', default=1000,
               help='Number of central greenthreads to spawn'),
    cfg.StrOpt('storage_driver', default='sqlalchemy',
               help='The storage driver to use'),
    cfg.IntOpt('max_zone_name_len', default=255,
               help="Maximum zone name length"),
    cfg.IntOpt('max_recordset_name_len', default=255,
               help="Maximum recordset name length",
               deprecated_name='max_record_name_len'),
    cfg.StrOpt('managed_resource_email', default='hostmaster@example.com',
               help='E-Mail for Managed resources'),
    cfg.StrOpt('managed_resource_tenant_id',
               default="00000000-0000-0000-0000-000000000000",
               help="The Tenant ID that will own any managed resources."),
    cfg.IntOpt('min_ttl', help="Minimum TTL allowed"),
    # TODO(betsy): Move to Pool Service once that is written
    cfg.StrOpt('default_pool_id',
               default='794ccc2c-d751-44fe-b57f-8894c9f5c842',
               help="The name of the default pool"),
    cfg.StrOpt('topic', default='central',
               help='RPC topic name for central'),
    cfg.ListOpt(
        'scheduler_filters',
        default=['default_pool'],
        help='Enabled Pool Scheduling filters'),
]


def register_opts(conf):
    conf.register_group(CENTRAL_GROUP)
    conf.register_opts(CENTRAL_OPTS, group=CENTRAL_GROUP)


def list_opts():
    return {
        CENTRAL_GROUP: CENTRAL_OPTS
    }



# Copyright (c) 2014 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from designate.objects import base
from designate.objects import fields
from designate.objects.record import Record
from designate.objects.record import RecordList


@base.DesignateRegistry.register
class SOA(Record):
    """
    SOA Resource Record Type
    Defined in: RFC1035
    """

    fields = {
        'mname': fields.DomainField(maxLength=255),
        'rname': fields.DomainField(maxLength=255),
        'serial': fields.IntegerFields(minimum=1, maximum=4294967295),
        'refresh': fields.IntegerFields(minimum=0, maximum=2147483647),
        'retry': fields.IntegerFields(minimum=0, maximum=2147483647),
        'expire': fields.IntegerFields(minimum=0, maximum=2147483647),
        'minimum': fields.IntegerFields(minimum=0, maximum=2147483647)
    }

    def from_string(self, value):
        mname, rname, serial, refresh, retry, expire, minimum = (
            value.split(' ')
        )
        self.mname = mname
        self.rname = rname
        self.serial = int(serial)
        self.refresh = int(refresh)
        self.retry = int(retry)
        self.expire = int(expire)
        self.minimum = int(minimum)

    # The record type is defined in the RFC. This will be used when the record
    # is sent by mini-dns.
    RECORD_TYPE = 6


@base.DesignateRegistry.register
class SOAList(RecordList):

    LIST_ITEM_TYPE = SOA

    fields = {
        'objects': fields.ListOfObjectsField('SOA'),
    }



# Copyright (c) 2014 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import binascii
from unittest import mock
from unittest.mock import patch

import dns
import dns.exception
import dns.message
import dns.query

from designate import objects
from designate.tests import TestCase
from designate.worker.tasks import zone


class WorkerNotifyTest(TestCase):
    test_zone = {
        'name': 'example.com.',
        'email': 'example@example.com',
        'serial': 100,
    }

    def setUp(self):
        super(WorkerNotifyTest, self).setUp()
        self.nameserver = objects.PoolNameserver.from_dict({
            'id': 'f278782a-07dc-4502-9177-b5d85c5f7c7e',
            'host': '127.0.0.1',
            'port': 65255
        })
        self.mock_tg = mock.Mock()

    def test_poll_for_serial_number(self):
        # id 10001
        # opcode QUERY
        # rcode NOERROR
        # flags QR AA
        # ;QUESTION
        # example.com. IN SOA
        # ;ANSWER
        # example.com. 3600 IN SOA example-ns.com. admin.example.com. 100 3600
        #  600 86400 3600
        # ;AUTHORITY
        # ;ADDITIONAL
        poll_response = ("271184000001000100000000076578616d706c6503636f6d0000"
                         "060001c00c0006000100000e1000290a6578616d706c652d6e73"
                         "c0140561646d696ec00c0000006400000e100000025800015180"
                         "00000e10")
        with patch.object(dns.query, 'udp', return_value=dns.message.from_wire(
                binascii.a2b_hex(poll_response))):
            get_zone_serial = zone.GetZoneSerial(
                self.mock_tg, 'context',
                objects.Zone.from_dict(self.test_zone),
                self.nameserver.host, self.nameserver.port,
            )
            result = get_zone_serial()
            self.assertEqual(result[0], 'SUCCESS')
            self.assertEqual(result[1], self.test_zone['serial'])

    def test_poll_for_serial_number_lower_serial(self):
        # id 10001
        # opcode QUERY
        # rcode NOERROR
        # flags QR AA
        # ;QUESTION
        # example.com. IN SOA
        # ;ANSWER
        # example.com. 3600 IN SOA example-ns.com. admin.example.com. 99 3600
        #  600 86400 3600
        # ;AUTHORITY
        # ;ADDITIONAL
        poll_response = ("271184000001000100000000076578616d706c6503636f6d0000"
                         "060001c00c0006000100000e1000290a6578616d706c652d6e73"
                         "c0140561646d696ec00c0000006300000e100000025800015180"
                         "00000e10")
        with patch.object(dns.query, 'udp', return_value=dns.message.from_wire(
                binascii.a2b_hex(poll_response))):
            get_zone_serial = zone.GetZoneSerial(
                self.mock_tg, 'context',
                objects.Zone.from_dict(self.test_zone),
                self.nameserver.host, self.nameserver.port,
            )
            result = get_zone_serial()
            self.assertEqual(result[0], 'SUCCESS')
            self.assertEqual(result[1], 99)

    def test_poll_for_serial_number_higher_serial(self):
        # id 10001
        # opcode QUERY
        # rcode NOERROR
        # flags QR AA
        # ;QUESTION
        # example.com. IN SOA
        # ;ANSWER
        # example.com. 3600 IN SOA example-ns.com. admin.example.com. 101 3600
        #  600 86400 3600
        # ;AUTHORITY
        # ;ADDITIONAL
        poll_response = ("271184000001000100000000076578616d706c6503636f6d0000"
                         "060001c00c0006000100000e1000290a6578616d706c652d6e73"
                         "c0140561646d696ec00c0000006500000e100000025800015180"
                         "00000e10")
        with patch.object(dns.query, 'udp', return_value=dns.message.from_wire(
                binascii.a2b_hex(poll_response))):
            get_zone_serial = zone.GetZoneSerial(
                self.mock_tg, 'context',
                objects.Zone.from_dict(self.test_zone),
                self.nameserver.host, self.nameserver.port,
            )
            result = get_zone_serial()
            self.assertEqual(result[0], 'SUCCESS')
            self.assertEqual(result[1], 101)

    @patch.object(dns.query, 'udp', side_effect=dns.exception.Timeout)
    def test_poll_for_serial_number_timeout(self, _):
        self.CONF.set_override('serial_timeout', 1, 'service:worker')
        get_zone_serial = zone.GetZoneSerial(
            self.mock_tg, 'context',
            objects.Zone.from_dict(self.test_zone),
            self.nameserver.host, self.nameserver.port,
        )
        result = get_zone_serial()
        self.assertEqual(result[0], 'ERROR')
        self.assertIsNone(result[1])



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""new_pools_tables

Revision ID: d04819112169
Revises: 0bcf910ea823
Create Date: 2022-08-01 16:20:17.440784

"""
from alembic import op
from oslo_utils import timeutils
import sqlalchemy as sa

from designate.sqlalchemy.types import UUID
from designate.storage.impl_sqlalchemy.alembic import legacy_utils
from designate import utils

# revision identifiers, used by Alembic.
revision = 'd04819112169'
down_revision = '0bcf910ea823'
branch_labels = None
depends_on = None

# Equivalent to legacy sqlalchemy-migrate revision 086_new_pools_tables


def upgrade() -> None:
    # Check if the equivalent legacy migration has already run
    if not legacy_utils.is_migration_needed(86):
        return

    metadata = sa.MetaData()

    op.create_table(
        'pool_nameservers', metadata,
        sa.Column('id', UUID, default=utils.generate_uuid, primary_key=True),
        sa.Column('version', sa.Integer, default=1, nullable=False),
        sa.Column('created_at', sa.DateTime,
                  default=lambda: timeutils.utcnow()),
        sa.Column('updated_at', sa.DateTime,
                  onupdate=lambda: timeutils.utcnow()),
        sa.Column('pool_id', UUID, nullable=False),
        sa.Column('host', sa.String(255), nullable=False),
        sa.Column('port', sa.Integer, nullable=False),
        sa.ForeignKeyConstraint(['pool_id'], ['pools.id'], ondelete='CASCADE'),
        sa.UniqueConstraint('pool_id', 'host', 'port',
                            name='unique_pool_host_port'))

    op.create_table(
        'pool_targets', metadata,
        sa.Column('id', UUID, default=utils.generate_uuid, primary_key=True),
        sa.Column('version', sa.Integer, default=1, nullable=False),
        sa.Column('created_at', sa.DateTime,
                  default=lambda: timeutils.utcnow()),
        sa.Column('updated_at', sa.DateTime,
                  onupdate=lambda: timeutils.utcnow()),
        sa.Column('pool_id', UUID, nullable=False),
        sa.Column('type', sa.String(50), nullable=False),
        sa.Column('tsigkey_id', UUID, nullable=True),
        sa.Column('description', sa.Unicode(160), nullable=True),
        sa.ForeignKeyConstraint(['pool_id'], ['pools.id'], ondelete='CASCADE'))

    op.create_table(
        'pool_target_masters', metadata,
        sa.Column('id', UUID, default=utils.generate_uuid, primary_key=True),
        sa.Column('version', sa.Integer, default=1, nullable=False),
        sa.Column('created_at', sa.DateTime,
                  default=lambda: timeutils.utcnow()),
        sa.Column('updated_at', sa.DateTime,
                  onupdate=lambda: timeutils.utcnow()),
        sa.Column('pool_target_id', UUID, nullable=False),
        sa.Column('host', sa.String(255), nullable=False),
        sa.Column('port', sa.Integer, nullable=False),
        sa.ForeignKeyConstraint(['pool_target_id'], ['pool_targets.id'],
                                ondelete='CASCADE'),
        sa.UniqueConstraint('pool_target_id', 'host', 'port',
                            name='unique_pool_target_host_port'))

    op.create_table(
        'pool_target_options', metadata,
        sa.Column('id', UUID, default=utils.generate_uuid, primary_key=True),
        sa.Column('version', sa.Integer, default=1, nullable=False),
        sa.Column('created_at', sa.DateTime,
                  default=lambda: timeutils.utcnow()),
        sa.Column('updated_at', sa.DateTime,
                  onupdate=lambda: timeutils.utcnow()),
        sa.Column('pool_target_id', UUID, nullable=False),
        sa.Column('key', sa.String(255), nullable=False),
        sa.Column('value', sa.String(255), nullable=False),
        sa.ForeignKeyConstraint(['pool_target_id'], ['pool_targets.id'],
                                ondelete='CASCADE'),
        sa.UniqueConstraint('pool_target_id', 'key',
                            name='unique_pool_target_key'))

    op.create_table(
        'pool_also_notifies', metadata,
        sa.Column('id', UUID, default=utils.generate_uuid, primary_key=True),
        sa.Column('version', sa.Integer, default=1, nullable=False),
        sa.Column('created_at', sa.DateTime,
                  default=lambda: timeutils.utcnow()),
        sa.Column('updated_at', sa.DateTime,
                  onupdate=lambda: timeutils.utcnow()),
        sa.Column('pool_id', UUID, nullable=False),
        sa.Column('host', sa.String(255), nullable=False),
        sa.Column('port', sa.Integer, nullable=False),
        sa.ForeignKeyConstraint(['pool_id'], ['pools.id'], ondelete='CASCADE'),
        sa.UniqueConstraint('pool_id', 'host', 'port',
                            name='unique_pool_also_notifies_pool0host0port'))



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import dns
import dns.flags
import dns.message
import dns.opcode
import dns.rcode
import dns.rdataclass
import dns.rdatatype
from oslo_config import cfg
from oslo_log import log as logging

from designate import exceptions
from designate.worker import rpcapi as worker_api


LOG = logging.getLogger(__name__)
CONF = cfg.CONF

CONF.import_opt('default_pool_id', 'designate.central',
                group='service:central')

# 10 Bytes of RR metadata, 64 bytes of TSIG RR data, variable length TSIG Key
# name (restricted in designate to 160 chars), 1 byte for trailing dot.
TSIG_RRSIZE = 10 + 64 + 160 + 1


class RequestHandler(object):
    def __init__(self, storage, tg):
        self._worker_api = None

        self.storage = storage
        self.tg = tg

    @property
    def worker_api(self):
        if not self._worker_api:
            self._worker_api = worker_api.WorkerAPI.get_instance()
        return self._worker_api

    def __call__(self, request):
        """
        :param request: DNS Request Message
        :return: DNS Response Message
        """
        if request.opcode() == dns.opcode.QUERY:
            # Currently we expect exactly 1 question in the section
            # TSIG places the pseudo records into the additional section.
            if (len(request.question) != 1 or
                    request.question[0].rdclass != dns.rdataclass.IN):
                LOG.debug('Refusing due to numbers of questions or rdclass')
                yield self._handle_query_error(request, dns.rcode.REFUSED)
                return

            q_rrset = request.question[0]
            # Handle AXFR and IXFR requests with an AXFR responses for now.
            # It is permissible for a server to send an AXFR response when
            # receiving an IXFR request.
            if q_rrset.rdtype in (dns.rdatatype.AXFR, dns.rdatatype.IXFR):
                for response in self._handle_axfr(request):
                    yield response
                return

            else:
                for response in self._handle_record_query(request):
                    yield response
                return

        elif request.opcode() == dns.opcode.NOTIFY:
            for response in self._handle_notify(request):
                yield response
            return

        else:
            # Unhandled OpCode's include STATUS, IQUERY, UPDATE
            LOG.debug('Refusing unhandled opcode')
            yield self._handle_query_error(request, dns.rcode.REFUSED)
            return

    def _handle_notify(self, request):
        """
        Constructs the response to a NOTIFY and acts accordingly on it.

        * Checks if the master sending the NOTIFY is in the Zone's masters,
          if not it is ignored.
        * Checks if SOA query response serial != local serial.
        """
        context = request.environ['context']

        response = dns.message.make_response(request)

        if len(request.question) != 1:
            response.set_rcode(dns.rcode.FORMERR)
            yield response
            return
        else:
            question = request.question[0]

        name = question.name.to_text()
        if isinstance(name, bytes):
            name = name.decode('utf-8')

        criterion = {
            'name': name,
            'type': 'SECONDARY',
            'deleted': False
        }

        try:
            zone = self.storage.find_zone(context, criterion)
        except exceptions.ZoneNotFound:
            response.set_rcode(dns.rcode.NOTAUTH)
            yield response
            return

        notify_addr = request.environ['addr'][0]

        # We check if the src_master which is the assumed master for the zone
        # that is sending this NOTIFY OP is actually the master. If it's not
        # We'll reply but don't do anything with the NOTIFY.
        master_addr = zone.get_master_by_ip(notify_addr)
        if not master_addr:
            LOG.warning(
                'NOTIFY for %(name)s from non-master server %(addr)s, '
                'refusing.',
                {
                    'name': zone.name,
                    'addr': notify_addr
                }
            )
            response.set_rcode(dns.rcode.REFUSED)
            yield response
            return

        resolver = dns.resolver.Resolver()
        # According to RFC we should query the server that sent the NOTIFY
        resolver.nameservers = [notify_addr]

        soa_answer = resolver.query(zone.name, 'SOA')
        soa_serial = soa_answer[0].serial

        if soa_serial == zone.serial:
            LOG.info(
                'Serial %(serial)s is the same for master and us for '
                '%(zone_id)s',
                {
                    'serial': soa_serial,
                    'zone_id': zone.id
                }
            )
        else:
            LOG.info(
                'Scheduling AXFR for %(zone_id)s from %(master_addr)s',
                {
                    'zone_id': zone.id,
                    'master_addr': master_addr.to_data()
                }
            )
            self.worker_api.perform_zone_xfr(context, zone, [master_addr])

        response.flags |= dns.flags.AA

        yield response
        return

    def _zone_criterion_from_request(self, request, criterion=None):
        """Builds a bare criterion dict based on the request attributes"""
        criterion = criterion or {}

        tsigkey = request.environ.get('tsigkey')

        if tsigkey is None and CONF['service:mdns'].query_enforce_tsig:
            raise exceptions.Forbidden('Request is not TSIG signed')
        elif tsigkey is None:
            # Default to using the default_pool_id when no TSIG key is
            # available
            criterion['pool_id'] = CONF['service:central'].default_pool_id
        else:
            if tsigkey.scope == 'POOL':
                criterion['pool_id'] = tsigkey.resource_id
            elif tsigkey.scope == 'ZONE':
                criterion['id'] = tsigkey.resource_id
            else:
                raise NotImplementedError('Support for %s scoped TSIG Keys is '
                                          'not implemented')
        return criterion

    def _handle_axfr(self, request):
        context = request.environ['context']
        q_rrset = request.question[0]

        # First check if there is an existing zone
        # TODO(vinod) once validation is separated from the api,
        # validate the parameters
        try:
            name = q_rrset.name.to_text()
            if isinstance(name, bytes):
                name = name.decode('utf-8')
            criterion = self._zone_criterion_from_request(
                request, {'name': name})
            zone = self.storage.find_zone(context, criterion)
        except exceptions.ZoneNotFound:
            LOG.warning('ZoneNotFound while handling axfr request. '
                        'Question was %(qr)s', {'qr': q_rrset})

            yield self._handle_query_error(request, dns.rcode.REFUSED)
            return
        except exceptions.Forbidden:
            LOG.warning('Forbidden while handling axfr request. '
                        'Question was %(qr)s', {'qr': q_rrset})

            yield self._handle_query_error(request, dns.rcode.REFUSED)
            return

        # The AXFR response needs to have a SOA at the beginning and end.
        criterion = {'zone_id': zone.id, 'type': 'SOA'}
        soa_records = self.storage.find_recordsets_axfr(context, criterion)

        # Get all the records other than SOA
        criterion = {'zone_id': zone.id, 'type': '!SOA'}
        records = self.storage.find_recordsets_axfr(context, criterion)

        # Place the SOA RRSet at the front and end of the RRSet list
        records.insert(0, soa_records[0])
        records.append(soa_records[0])

        # Handle multi message response with tsig
        multi_messages = False
        multi_messages_context = None

        # Render the results, yielding a packet after each TooBig exception.
        renderer = None
        while records:
            record = records.pop(0)

            rrname = str(record[3])
            ttl = int(record[2]) if record[2] is not None else zone.ttl
            rrtype = str(record[1])
            rdata = [str(record[4])]

            rrset = dns.rrset.from_text_list(
                rrname, ttl, dns.rdataclass.IN, rrtype, rdata,
            )

            while True:
                try:
                    if not renderer:
                        renderer = self._create_axfr_renderer(request)
                    renderer.add_rrset(dns.renderer.ANSWER, rrset)
                    break
                except dns.exception.TooBig:
                    # The response will span multiple messages since one
                    # message is not enough
                    multi_messages = True
                    if renderer.counts[dns.renderer.ANSWER] == 0:
                        # We've received a TooBig from the first attempted
                        # RRSet in this packet. Log a warning and abort the
                        # AXFR.
                        LOG.warning(
                            'Aborted AXFR of %(zone)s, a single RR '
                            '(%(rrset_type)s %(rrset_name)s) '
                            'exceeded the max message size.',
                            {
                                'zone': zone.name,
                                'rrset_type': rrtype,
                                'rrset_name': rrname,
                            }
                        )

                        yield self._handle_query_error(
                            request, dns.rcode.SERVFAIL
                        )
                        return

                    renderer, multi_messages_context = self._finalize_packet(
                        renderer, request, multi_messages,
                        multi_messages_context)
                    yield renderer
                    renderer = None

        if renderer:
            renderer, multi_messages_context = self._finalize_packet(
                renderer, request, multi_messages, multi_messages_context)
            yield renderer
        return

    def _handle_record_query(self, request):
        """Handle a DNS QUERY request for a record"""
        context = request.environ['context']
        response = dns.message.make_response(request)

        try:
            q_rrset = request.question[0]
            name = q_rrset.name.to_text()
            if isinstance(name, bytes):
                name = name.decode('utf-8')
            # TODO(vinod) once validation is separated from the api,
            # validate the parameters
            criterion = {
                'name': name,
                'type': dns.rdatatype.to_text(q_rrset.rdtype),
                'zones_deleted': False
            }
            recordset = self.storage.find_recordset(context, criterion)

        except exceptions.NotFound:
            # If an FQDN exists, like www.rackspace.com, but the specific
            # record type doesn't exist, like type SPF, then the return code
            # would be NOERROR and the SOA record is returned.  This tells
            # caching nameservers that the FQDN does exist, so don't negatively
            # cache it, but the specific record doesn't exist.
            #
            # If an FQDN doesn't exist with any record type, that is NXDOMAIN.
            # However, an authoritative nameserver shouldn't return NXDOMAIN
            # for a zone it isn't authoritative for.  It would be more
            # appropriate for it to return REFUSED.  It should still return
            # NXDOMAIN if it is authoritative for a zone but the FQDN doesn't
            # exist, like abcdef.rackspace.com.  Of course, a wildcard within a
            # zone would mean that NXDOMAIN isn't ever returned for a zone.
            #
            # To simply things currently this returns a REFUSED in all cases.
            # If zone transfers needs different errors, we could revisit this.
            LOG.info('NotFound, refusing. Question was %(qr)s',
                     {'qr': q_rrset})
            yield self._handle_query_error(request, dns.rcode.REFUSED)
            return

        except exceptions.Forbidden:
            LOG.info('Forbidden, refusing. Question was %(qr)s',
                     {'qr': q_rrset})
            yield self._handle_query_error(request, dns.rcode.REFUSED)
            return

        try:
            criterion = self._zone_criterion_from_request(
                request, {'id': recordset.zone_id})
            zone = self.storage.find_zone(context, criterion)

        except exceptions.ZoneNotFound:
            LOG.warning('ZoneNotFound while handling query request. '
                        'Question was %(qr)s', {'qr': q_rrset})
            yield self._handle_query_error(request, dns.rcode.REFUSED)
            return

        except exceptions.Forbidden:
            LOG.warning('Forbidden while handling query request. '
                        'Question was %(qr)s', {'qr': q_rrset})
            yield self._handle_query_error(request, dns.rcode.REFUSED)
            return

        r_rrset = self._convert_to_rrset(zone, recordset)
        response.answer = [r_rrset] if r_rrset else []
        response.set_rcode(dns.rcode.NOERROR)
        # For all the data stored in designate mdns is Authoritative
        response.flags |= dns.flags.AA
        yield response

    def _create_axfr_renderer(self, request):
        # Build up a dummy response, we're stealing it's logic for building
        # the Flags.
        response = dns.message.make_response(request)
        response.flags |= dns.flags.AA
        response.set_rcode(dns.rcode.NOERROR)

        max_message_size = self._get_max_message_size(request.had_tsig)

        renderer = dns.renderer.Renderer(
            response.id, response.flags, max_message_size)
        for q in request.question:
            renderer.add_question(q.name, q.rdtype, q.rdclass)
        return renderer

    @staticmethod
    def _convert_to_rrset(zone, recordset):
        # Fetch the zone or the config ttl if the recordset ttl is null
        ttl = recordset.ttl or zone.ttl

        # construct rdata from all the records
        # TODO(Ron): this should be handled in the Storage query where we
        # find the recordsets.
        rdata = [str(record.data) for record in recordset.records
                 if record.action != 'DELETE']

        # Now put the records into dnspython's RRsets
        # answer section has 1 RR set.  If the RR set has multiple
        # records, DNSpython puts each record in a separate answer
        # section.
        # RRSet has name, ttl, class, type  and rdata
        # The rdata has one or more records
        if not rdata:
            return None
        return dns.rrset.from_text_list(
            recordset.name, ttl, dns.rdataclass.IN, recordset.type, rdata)

    @staticmethod
    def _finalize_packet(renderer, request, multi_messages=False,
                         multi_messages_context=None):
        renderer.write_header()
        if request.had_tsig:
            # Make the space we reserved for TSIG available for use
            renderer.max_size += TSIG_RRSIZE
            if multi_messages:
                # The first message context will be None then the
                # context for the prev message is used for the next
                multi_messages_context = renderer.add_multi_tsig(
                    multi_messages_context, request.keyname,
                    request.keyring[request.keyname], request.fudge,
                    request.original_id, request.tsig_error,
                    request.other_data, request.mac, request.keyalgorithm)
            else:
                renderer.add_tsig(request.keyname,
                    request.keyring[request.keyname], request.fudge,
                    request.original_id, request.tsig_error,
                    request.other_data, request.mac, request.keyalgorithm)
        return renderer, multi_messages_context

    @staticmethod
    def _get_max_message_size(had_tsig):
        max_message_size = CONF['service:mdns'].max_message_size
        if max_message_size > 65535:
            LOG.warning('MDNS max message size must not be greater than 65535')
            max_message_size = 65535
        if had_tsig:
            # Make some room for the TSIG RR to be appended at the end of the
            # rendered message.
            max_message_size = max_message_size - TSIG_RRSIZE
        return max_message_size

    @staticmethod
    def _handle_query_error(request, rcode):
        """
        Construct an error response with the rcode passed in.
        :param request: The decoded request from the wire.
        :param rcode: The response code to send back.
        :return: A dns response message with the response code set to rcode
        """
        response = dns.message.make_response(request)
        response.set_rcode(rcode)

        return response



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add_zone_attributes

Revision ID: 0bcf910ea823
Revises: a69b45715cc1
Create Date: 2022-07-29 21:36:15.117658

"""
from alembic import op
from oslo_utils import timeutils
import sqlalchemy as sa

from designate.sqlalchemy.types import UUID
from designate.storage.impl_sqlalchemy.alembic import legacy_utils
from designate import utils

# revision identifiers, used by Alembic.
revision = '0bcf910ea823'
down_revision = 'a69b45715cc1'
branch_labels = None
depends_on = None

# Equivalent to legacy sqlalchemy-migrate revision 085_add_zone_attributes

# Note from original migration file:
# Move zone masters to their own table, and allow for abstract keys in the
# attributes table


def upgrade() -> None:
    # Check if the equivalent legacy migration has already run
    if not legacy_utils.is_migration_needed(85):
        return

    metadata = sa.MetaData()

    zone_masters_table = op.create_table(
        'zone_masters', metadata,
        sa.Column('id', UUID, default=utils.generate_uuid, primary_key=True),
        sa.Column('version', sa.Integer, default=1, nullable=False),
        sa.Column('created_at', sa.DateTime,
                  default=lambda: timeutils.utcnow()),
        sa.Column('updated_at', sa.DateTime,
                  onupdate=lambda: timeutils.utcnow()),
        sa.Column('host', sa.String(32), nullable=False),
        sa.Column('port', sa.Integer, nullable=False),
        sa.Column('zone_id', UUID, nullable=False),
        sa.UniqueConstraint('host', 'port', 'zone_id', name='unique_masters'),
        sa.ForeignKeyConstraint(['zone_id'], ['zones.id'], ondelete='CASCADE'))

    zone_attr_sql = sa.text(
        'SELECT id, version, created_at, updated_at, value, zone_id FROM '
        'zone_attributes WHERE \'key\' = \'master\';')

    masters = op.get_bind().execute(zone_attr_sql).fetchall()

    masters_input = []

    for master in masters:
        host, port = utils.split_host_port(
            master['value'])
        masters_input.append({
            'id': master['id'],
            'version': master['version'],
            'created_at': master['created_at'],
            'updated_at': master['updated_at'],
            'zone_id': master['zone_id'],
            'host': host,
            'port': port
        })

    op.bulk_insert(zone_masters_table, masters_input)

    master_delete_sql = sa.text(
        'DELETE FROM zone_attributes WHERE \'key\' = \'master\';')

    op.get_bind().execute(master_delete_sql)

    with op.batch_alter_table('zone_attributes') as batch_op:
        batch_op.alter_column('key', type_=sa.String(50),
                              existing_type=sa.Enum)
